class ConfidenceInterval:
    """
    A class to store confidence interval values.
    """

    def __init__(self, lower: float, upper: float) -> None:
        self.lower: float = float(lower)
        self.upper: float = float(upper)
        self.ci: tuple[float, float] = (self.lower, self.upper)


class ApproximatedStandardError:
    """
    A class to store approximated standard error values.
    """

    def __init__(
        self,
        true_se: float,
        morris: float,
        hedges: float,
        hedges_olkin: float,
        mle: float,
        large_n: float,
        hunter_and_schmidt: float,
    ) -> None:
        self.true_se: float = true_se
        self.morris: float = morris
        self.hedges: float = hedges
        self.hedges_olkin: float = hedges_olkin
        self.mle: float = mle
        self.large_n: float = large_n
        self.hunter_and_schmidt: float = hunter_and_schmidt


class EffectSize:
    """
    A class to store effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        self.effect_size_name: str = "Effect Size"
        self.value: float = value
        self.ci: ConfidenceInterval = ConfidenceInterval(ci_lower, ci_upper)
        self.standard_error: float = standard_error
        self.statistical_line: str = ""

        self.standardizer: float | None = None
        self.non_central_ci: ConfidenceInterval | None = None
        self.pivotal_ci: ConfidenceInterval | None = None
        self.approximated_standard_error: ApproximatedStandardError | None = None
        self.mag: ConfidenceInterval | None = None
        self.lambda_prime: ConfidenceInterval | None = None
        self.morris: ConfidenceInterval | None = None
        self.t_prime: ConfidenceInterval | None = None

    def update_statistical_line(self) -> None:
        """
        Update the statistical line with the current values.
        """
        self.statistical_line = (
            f"{self.effect_size_name}: {self.value} (CI: {self.ci.lower}, {self.ci.upper}) "
            f"SE: {self.standard_error}"
        )

        if self.non_central_ci is not None:
            self.statistical_line += f" Non-Central CI: {self.non_central_ci.lower}, {self.non_central_ci.upper}"

        if self.pivotal_ci is not None:
            self.statistical_line += (
                f" Pivotal CI: {self.pivotal_ci.lower}, {self.pivotal_ci.upper}"
            )

    def update_non_central_ci(
        self, non_central_ci_lower: float, non_central_ci_upper: float
    ) -> None:
        """
        Update the non-central confidence interval.
        """
        self.non_central_ci = ConfidenceInterval(
            non_central_ci_lower, non_central_ci_upper
        )
        self.update_statistical_line()

    def update_pivotal_ci(
        self, pivotal_ci_lower: float, pivotal_ci_upper: float
    ) -> None:
        """
        Update the pivotal confidence interval.
        """
        self.pivotal_ci = ConfidenceInterval(pivotal_ci_lower, pivotal_ci_upper)
        self.update_statistical_line()

    def update_mag_ci(self, mag_ci_lower: float, mag_ci_upper: float) -> None:
        """
        Update the mag confidence interval.
        """
        self.mag = ConfidenceInterval(mag_ci_lower, mag_ci_upper)
        self.update_statistical_line()

    def update_lambda_prime_ci(
        self, lambda_prime_ci_lower: float, lambda_prime_ci_upper: float
    ) -> None:
        """
        Update the lambda prime confidence interval.
        """
        self.lambda_prime = ConfidenceInterval(
            lambda_prime_ci_lower, lambda_prime_ci_upper
        )
        self.update_statistical_line()

    def update_morris_ci(self, morris_ci_lower: float, morris_ci_upper: float) -> None:
        """
        Update the morris prime confidence interval.
        """
        self.morris = ConfidenceInterval(morris_ci_lower, morris_ci_upper)
        self.update_statistical_line()

    def update_t_prime_ci(
        self, t_prime_ci_lower: float, t_prime_ci_upper: float
    ) -> None:
        """
        Update the t prime prime confidence interval.
        """
        self.t_prime = ConfidenceInterval(t_prime_ci_lower, t_prime_ci_upper)
        self.update_statistical_line()


class CohenD(EffectSize):
    """
    A class to store Cohen's d effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cohen's d"
        self.update_statistical_line()


class HedgesG(EffectSize):
    """
    A class to store Hedges' g effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Hedges' g"
        self.update_statistical_line()


class CohensDav(EffectSize):
    """
    A class to store Cohen's dav effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cohen's dav"
        self.update_statistical_line()


class HedgesGav(EffectSize):
    """
    A class to store Hedge's gav effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Hedge's gav"
        self.update_statistical_line()


class CohensDrm(EffectSize):
    """
    A class to store Cohen's drm effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cohen's drm"
        self.update_statistical_line()


class HedgesGrm(EffectSize):
    """
    A class to store Hedge's grm effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Hedge's grm"
        self.update_statistical_line()


class Biserial(EffectSize):
    """
    A class to store Biserial effect size values.
    """

    def __init__(
        self,
        value: float,
        ci_lower: float,
        ci_upper: float,
        standard_error: float,
        name: str,
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = name
        self.z_score: float | None = None
        self.p_value: float | None = None
        self.update_statistical_line()


class RatioOfMeans(EffectSize):
    """
    A class to store Ratio of Means effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Ratio of Means"
        self.update_statistical_line()


class RobustAKP(EffectSize):
    """
    A class to store Robust AKP effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Robust AKP"
        self.update_statistical_line()


class RobustExplanatory(EffectSize):
    """
    A class to store Robust Explanatory effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Robust Explanatory"
        self.update_statistical_line()


class CLES(EffectSize):
    """
    A class to store Common Language Effect Size (CLES) values.
    """

    def __init__(
        self,
        value: float,
        ci_lower: float,
        ci_upper: float,
        standard_error: float,
        method: str,
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = (
            "Common Language Effect Size (CLES)" + f" ({method})"
        )
        self.update_statistical_line()


class ProbabilityOfSuperiority(EffectSize):
    """
    A class to store Probability of Superiority (PS) values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Probability of Superiority (PS)"
        self.update_statistical_line()


class VarghaDelaney(EffectSize):
    """
    A class to store Vargha-Delaney effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Vargha-Delaney"
        self.update_statistical_line()


class CliffsDelta(EffectSize):
    """
    A class to store Cliff's Delta effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cliff's Delta"
        self.update_statistical_line()


class NonParametricU1(EffectSize):
    """
    A class to store Non-Parametric U1 effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Non-Parametric U1"
        self.update_statistical_line()


class NonParametricU3(EffectSize):
    """
    A class to store Non-Parametric U3 effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Non-Parametric U3"
        self.update_statistical_line()


class KraemerAndrewGamma(EffectSize):
    """
    A class to store Kraemer-Andrew Gamma effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Kraemer-Andrew Gamma"
        self.update_statistical_line()


class WilcoxMusakaQ(EffectSize):
    """
    A class to store Wilcox-Musaka Q effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Wilcox-Musaka Q"
        self.update_statistical_line()


class InferentialStatistics:
    """
    A class to store inferential statistics.
    """

    def __init__(
        self,
        p_value: float,
        score: float,
    ) -> None:
        self.p_value: float = p_value
        self.score: float = score
        self.standard_error: float | None = None
        self.degrees_of_freedom: float | None = None
        self.means_difference: float | None = None


class DescriptiveStatistics:
    """
    A class to store descriptive statistics.
    """

    def __init__(
        self,
        mean: float,
        standard_deviation: float,
    ) -> None:
        self.mean: float = mean
        self.sd: float = standard_deviation


class Sample(DescriptiveStatistics):
    """
    A class to store descriptive statistics for a sample.
    """

    def __init__(
        self,
        mean: float,
        standard_deviation: float,
        size: int,
    ) -> None:
        super().__init__(mean, standard_deviation)
        self.size: int = size
        self.diff_mean: float | None = None
        self.diff_sd: float | None = None
        self.population_sd_diff: float | None = None
        self.population_mean: float | None = None


class Group(DescriptiveStatistics):
    """
    A class to store descriptive statistics for a group.
    """

    def __init__(
        self,
        mean: float,
        standard_deviation: float,
        median: float | None = None,
        median_absolute_deviation: float | None = None,
    ) -> None:
        super().__init__(mean, standard_deviation)
        self.median: float | None = median
        self.median_absolute_deviation: float | None = median_absolute_deviation
        self.diff_median: float | None = None
        self.sample_size: int | None = None
        self.u_statistic: float | None = None
        self.w_statistic: float | None = None
        self.mean_rank: float | None = None
        self.population_sd: float | None = None
        self.mean_diff: float | None = None
        self.sd_diff: float | None = None


class WilcoxonSignedRank(DescriptiveStatistics):
    """
    A class to store descriptive statistics from the Wilcoxon Signed-Rank test.
    """

    def __init__(
        self,
        times_group1_larger: float,
        times_group2_larger: float,
        ties: float,
        num_of_pairs: float,
        num_of_non_tied_pairs: float,
    ) -> None:
        self.times_group1_larger: float = times_group1_larger
        self.times_group2_larger: float = times_group2_larger
        self.ties: float = ties
        self.num_of_pairs: float = num_of_pairs
        self.num_of_non_tied_pairs: float = num_of_non_tied_pairs



from src.esek.Calculator.TwoPairedMean.two_paired_common_lang import (
    TwoPairedCommonLangResults,
    TwoPairedCommonLangTests,
)


def test_two_paired_common_lang_from_score():
    data = {
        "t_score": 1.96,
        "sample_size": 30,
        "confidence_level": 0.95,
    }
    try:
        results = TwoPairedCommonLangTests.from_score(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, TwoPairedCommonLangResults)


def test_two_paired_common_lang_from_parameters():
    data = {
        "sample_mean_1": 100,
        "sample_mean_2": 75,
        "sample_sd_1": 10,
        "sample_sd_2": 7.5,
        "sample_size": 30,
        "population_mean_diff": 5,
        "correlation": 0.8,
        "confidence_level": 0.95,
    }
    try:
        results = TwoPairedCommonLangTests.from_parameters(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, TwoPairedCommonLangResults)


def test_two_paired_common_lang_from_data():
    assert True, "Return division by zero error"
    return
    data = {
        "column": [[1, 2, 3], [4, 5, 6]],
        "reps": 1000,
        "confidence_level": 0.95,
    }
    try:
        results = TwoPairedCommonLangTests.from_data(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, TwoPairedCommonLangResults)



from src.esek.Calculator.OneSampleMean.one_sample_z import (
    OneSampleZResults,
    OneSampleZTests,
)


def test_one_sample_z_from_score():
    data = {
        "z_score": 1.96,
        "sample_size": 30,
        "confidence_level": 0.95,
    }
    try:
        results = OneSampleZTests.from_score(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, OneSampleZResults)


def test_one_sample_z_from_parameters():
    data = {
        "sample_mean": 100,
        "sample_size": 30,
        "population_mean": 95,
        "population_sd": 0.5,
        "confidence_level": 0.95,
    }
    try:
        results = OneSampleZTests.from_parameters(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, OneSampleZResults)


def test_one_sample_z_from_data():
    data = {
        "column": [[1, 2, 3], [4, 5, 6]],
        "population_mean": 95,
        "population_sd": 0.5,
        "confidence_level": 0.95,
    }
    try:
        results = OneSampleZTests.from_data(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, OneSampleZResults)



from src.esek.Calculator.TwoIndependentMean.two_independent_aparametric import (
    TwoIndependentAparametricResults,
    TwoIndependentAparametricTests,
)


def test_two_independent_aparametric_from_score():
    assert True, "Method not implemented yet"


def test_two_independent_aparametric_from_parameters():
    assert True, "Method not implemented yet"


def test_two_independent_aparametric_from_data():
    data = {
        "columns": [[1, 2, 3], [4, 5, 6]],
        "confidence_level": 0.95,
    }
    try:
        results = TwoIndependentAparametricTests.from_data(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, TwoIndependentAparametricResults)



from .results import EffectSize


class CohenD(EffectSize):
    """
    A class to store Cohen's d effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cohen's d"
        self.update_statistical_line()


class HedgesG(EffectSize):
    """
    A class to store Hedges' g effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Hedges' g"
        self.update_statistical_line()


class CohensDav(EffectSize):
    """
    A class to store Cohen's dav effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cohen's dav"
        self.update_statistical_line()


class HedgesGav(EffectSize):
    """
    A class to store Hedge's gav effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Hedge's gav"
        self.update_statistical_line()


class CohensDrm(EffectSize):
    """
    A class to store Cohen's drm effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cohen's drm"
        self.update_statistical_line()


class HedgesGrm(EffectSize):
    """
    A class to store Hedge's grm effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Hedge's grm"
        self.update_statistical_line()


class CohensDPop(EffectSize):
    """
    A class to store Cohen's d pop effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cohen's d pop"
        self.update_statistical_line()


class Biserial(EffectSize):
    """
    A class to store Biserial effect size values.
    """

    def __init__(
        self,
        value: float,
        ci_lower: float,
        ci_upper: float,
        standard_error: float,
        name: str,
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = name
        self.z_score: float | None = None
        self.p_value: float | None = None
        self.update_statistical_line()


class RatioOfMeans(EffectSize):
    """
    A class to store Ratio of Means effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Ratio of Means"
        self.update_statistical_line()


class RobustAKP(EffectSize):
    """
    A class to store Robust AKP effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Robust AKP"
        self.update_statistical_line()


class RobustExplanatory(EffectSize):
    """
    A class to store Robust Explanatory effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Robust Explanatory"
        self.update_statistical_line()


class CLES(EffectSize):
    """
    A class to store Common Language Effect Size (CLES) values.
    """

    def __init__(
        self,
        value: float,
        ci_lower: float,
        ci_upper: float,
        standard_error: float,
        method: str,
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = (
            "Common Language Effect Size (CLES)" + f" ({method})"
        )
        self.update_statistical_line()


class ProbabilityOfSuperiority(EffectSize):
    """
    A class to store Probability of Superiority (PS) values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Probability of Superiority (PS)"
        self.update_statistical_line()


class VarghaDelaney(EffectSize):
    """
    A class to store Vargha-Delaney effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Vargha-Delaney"
        self.update_statistical_line()


class CliffsDelta(EffectSize):
    """
    A class to store Cliff's Delta effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Cliff's Delta"
        self.update_statistical_line()


class NonParametricU1(EffectSize):
    """
    A class to store Non-Parametric U1 effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Non-Parametric U1"
        self.update_statistical_line()


class NonParametricU3(EffectSize):
    """
    A class to store Non-Parametric U3 effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Non-Parametric U3"
        self.update_statistical_line()


class KraemerAndrewGamma(EffectSize):
    """
    A class to store Kraemer-Andrew Gamma effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Kraemer-Andrew Gamma"
        self.update_statistical_line()


class WilcoxMusakaQ(EffectSize):
    """
    A class to store Wilcox-Musaka Q effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Wilcox-Musaka Q"
        self.update_statistical_line()


class GlassDelta(EffectSize):
    """
    A class to store Glass's Delta effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Glass's Delta"
        self.update_statistical_line()


class GlassDeltaUnbiased(EffectSize):
    """
    A class to store Glass's Delta Unbiased effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Glass's Delta Unbiased"
        self.update_statistical_line()


class AokiEpsilon(EffectSize):
    """
    A class to store Aoki's Epsilon effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Aoki's Epsilon"
        self.update_statistical_line()


class AokiEpsilonUnbiased(EffectSize):
    """
    A class to store Aoki's Epsilon Unbiased effect size values.
    """

    def __init__(
        self, value: float, ci_lower: float, ci_upper: float, standard_error: float
    ) -> None:
        super().__init__(value, ci_lower, ci_upper, standard_error)
        self.effect_size_name: str = "Aoki's Epsilon Unbiased"
        self.update_statistical_line()



"""
This module implements the Two Paired Z-test for comparing two paired samples.
It provides methods to calculate the test results from a Z-score, sample parameters,
or sample data. The results include Cohen's d, z-score, p-value, and confidence intervals for the effect size.
The TwoPairedZResults class encapsulates the results of the test and includes attributes such as
sample means, standard deviations, and sample sizes for both groups.
It is designed to be used in statistical analysis where paired samples are compared,
such as in pre-test and post-test scenarios.
"""

from dataclasses import dataclass
from typing import Optional
import numpy as np
from scipy import stats
from ...utils import interfaces, res, utils, es


@dataclass
class TwoPairedZResults:
    """
    a dataclass for the results of a two paired Z-test.
    """

    cohens_d: Optional[es.CohenD] = None
    inferential: Optional[res.InferentialStatistics] = None
    sample1: Optional[res.Sample] = None
    sample2: Optional[res.Sample] = None


class TwoPairedZTests(interfaces.AbstractTest):
    """
    A class for performing two paired Z-tests.
    This class provides methods to calculate Z-test results from a Z-score,
    from sample parameters, and from sample data.
    It encapsulates the results in a TwoPairedZResults object, which includes
    Cohen's d, z-score, p-value, and confidence intervals for the effect size.

    Methods:
        - from_z_score: Calculate the two paired Z-test results from a given Z-score.
        - from_parameters: Calculate the two paired Z-test results from sample parameters.
        - from_data: Calculate the two paired Z-test results from given sample data.
    """

    @staticmethod
    def from_score(
        z_score: float, sample_size: float, confidence_level: float
    ) -> TwoPairedZResults:
        """
        Calculate the two paired Z-test results from a given Z-score.
        This method computes the effect size (Cohen's d), p-value, and confidence intervals
        for the effect size based on the provided Z-score and sample size.
        It returns an instance of TwoPairedZResults containing these values.
        """

        p_value = min(float(stats.norm.sf((abs(z_score))) * 2), 0.99999)
        cohens_d = z_score / np.sqrt(sample_size)
        ci_lower, ci_upper, standard_error_es = utils.ci_from_cohens_simple(
            cohens_d, sample_size, confidence_level
        )
        cohens_d = es.CohenD(
            value=round(cohens_d, 4),
            ci_lower=round(ci_lower, 4),
            ci_upper=round(ci_upper, 4),
            standard_error=round(standard_error_es, 4),
        )

        inferential = res.InferentialStatistics(
            p_value=round(p_value, 4),
            score=round(z_score, 4),
        )

        results = TwoPairedZResults()
        results.cohens_d = cohens_d
        results.inferential = inferential

        return results

    @staticmethod
    def from_parameters(
        sample_mean_1: float,
        sample_mean_2: float,
        sample_size: float,
        population_diff: float,
        population_diff_sd: float,
        confidence_level: float,
    ) -> TwoPairedZResults:
        """
        Calculate the two paired Z-test results from sample parameters.
        This method computes the effect size (Cohen's d), z-score, p-value,
        and confidence intervals for the effect size based on the provided sample means,
        sample size, population difference, and population standard deviation of the difference.
        It returns an instance of TwoPairedZResults containing these values.
        """
        mean_Standard_error = population_diff_sd / np.sqrt(sample_size)
        sample_mean_diff = sample_mean_1 - sample_mean_2
        z_score = (sample_mean_diff - population_diff) / mean_Standard_error
        cohens_d = (sample_mean_diff - population_diff) / population_diff_sd
        p_value = min(float(stats.norm.sf((abs(z_score))) * 2), 0.99999)
        ci_lower, ci_upper, standard_error_es = utils.ci_from_cohens_simple(
            cohens_d, sample_size, confidence_level
        )

        cohens_d = es.CohenD(
            value=round(cohens_d, 4),
            ci_lower=round(ci_lower, 4),
            ci_upper=round(ci_upper, 4),
            standard_error=round(standard_error_es, 4),
        )

        inferential = res.InferentialStatistics(
            p_value=round(p_value, 4),
            score=round(z_score, 4),
        )
        inferential.standard_error = round(mean_Standard_error, 4)

        results = TwoPairedZResults()
        results.cohens_d = cohens_d
        results.inferential = inferential

        return results

    @staticmethod
    def from_data(
        columns: list,
        population_diff: float,
        population_diff_sd: float,
        confidence_level: float,
    ) -> TwoPairedZResults:
        """
        Calculate the two paired Z-test results from given sample data.
        This method computes the effect size (Cohen's d), z-score, p-value,
        and confidence intervals for the effect size based on the provided sample data,
        population difference, and population standard deviation of the difference.
        It returns an instance of TwoPairedZResults containing these values.
        """
        sample_mean_1 = np.mean(columns[0])
        sample_mean_2 = np.mean(columns[1])
        sample_sd_1 = np.std(columns[0], ddof=1)
        sample_sd_2 = np.std(columns[1], ddof=1)

        diff_mean = (sample_mean_1 - sample_mean_2) - population_diff

        sample_size = len(columns[0])
        standard_error = population_diff_sd / (np.sqrt(sample_size))

        z_score = diff_mean / standard_error
        cohens_d = (diff_mean) / population_diff_sd
        p_value = min(float(stats.norm.sf((abs(z_score))) * 2), 0.99999)
        ci_lower, ci_upper, standard_error_es = utils.ci_from_cohens_simple(
            cohens_d, sample_size, confidence_level
        )

        cohens_d = es.CohenD(
            value=round(cohens_d, 4),
            ci_lower=round(ci_lower, 4),
            ci_upper=round(ci_upper, 4),
            standard_error=round(standard_error_es, 4),
        )

        inferential = res.InferentialStatistics(
            p_value=round(p_value, 4),
            score=round(z_score, 4),
        )
        inferential.standard_error = round(standard_error, 4)
        sample1 = res.Sample(
            mean=round(sample_mean_1, 4),
            standard_deviation=round(float(sample_sd_1), 4),
            size=round(sample_size, 4),
        )
        sample2 = res.Sample(
            mean=round(sample_mean_2, 4),
            standard_deviation=round(float(sample_sd_2), 4),
            size=round(sample_size, 4),
        )
        samples: list[res.Sample] = [sample1, sample2]
        for sample in samples:
            sample.diff_mean = round(diff_mean, 4)
            sample.diff_sd = round(population_diff_sd, 4)
            sample.population_sd_diff = round(population_diff_sd, 4)

        results = TwoPairedZResults()
        results.cohens_d = cohens_d
        results.inferential = inferential
        results.sample1 = sample1
        results.sample2 = sample2

        return results



"""
Two Independent Robust Tests Module
This module provides functionality for conducting two independent robust tests,
including the calculation of various robust statistics and effect sizes.
"""

from dataclasses import dataclass
from typing import Optional
import numpy as np
from scipy import stats
from ...utils import interfaces, res, utils, es, texts


@dataclass
class TwoIndependentRobustResults:
    """
    This class holds the results of two independent robust tests,
    including the samples, inferential statistics, and robust effect sizes.
    """

    sample1: Optional[res.Sample] = None
    sample2: Optional[res.Sample] = None
    inferential: Optional[res.InferentialStatistics] = None
    robust_akp: Optional[es.RobustAKP] = None
    robust_exp: Optional[es.RobustExplanatory] = None


class TwoIndependentRobustTests(interfaces.AbstractTest):
    """
    This class implements two independent robust tests.
    It provides methods to calculate robust effect sizes and inferential statistics.

        It is designed to handle two independent samples and compute various statistics.

        Attributes:
                None

        Methods:
                - from_score: Create an instance from a score (Not implemented).
                - from_parameters: Create an instance from parameters (Not implemented).
                - from_data: Create an instance from data.
    """

    @utils.not_implemented(
        interfaces.MethodType.FROM_SCORE, "TwoIndependentRobustTests"
    )
    @staticmethod
    def from_score() -> None:
        """
        A static method to create results from a score.
        This method is not implemented and will raise a NotImplementedError.
        Raises:
                NotImplementedError: This method is not implemented for TwoIndependentRobustTests.
        """
        pass

    @utils.not_implemented(
        interfaces.MethodType.FROM_PARAMETERS, "TwoIndependentRobustTests"
    )
    @staticmethod
    def from_parameters() -> None:
        """
        A static method to create results from parameters.
        This method is not implemented and will raise a NotImplementedError.
        Raises:
                NotImplementedError: This method is not implemented for TwoIndependentRobustTests
        """
        pass

    @staticmethod
    def from_data(
        columns: list[list],
        reps: int = 10000,
        trimming_level: float = 0.2,
        population_difference: float = 0.2,
        confidence_level: float = 0.95,
    ) -> TwoIndependentRobustResults:
        """
        Create an instance from data.

        Args:
            columns (list[list]): The data columns.
            reps (int): The number of bootstrap repetitions. Default is 10000.
            trimming_level (float): The trimming level for robust statistics. Default is 0.2.
            population_difference (float): The population difference for effect size calculation. Default is 0.2.
            confidence_level (float): The confidence level for interval estimation. Default is 0.95.

        Returns:
            TwoIndependentRobustResults: The results of the robust tests.

        Raises:
            ValueError: If the input data is not valid.
        """

        if len(columns) != 2:
            raise ValueError(texts.Errors.columns_must_be_two)
        column_1 = columns[0]
        column_2 = columns[1]
        sample_size_1 = len(column_1)
        sample_size_2 = len(column_1)
        sample_size = sample_size_1 + sample_size_2
        lower_bound = stats.norm.ppf(trimming_level)
        upper_bound = stats.norm.ppf(1 - trimming_level)

        area = utils.area_under_function(
            utils.density,
            float(lower_bound),
            float(upper_bound),
        )

        tail_correction = 2 * (lower_bound**2) * trimming_level
        correction = np.sqrt(area + tail_correction)

        trimmed_mean_1 = stats.trim_mean(column_1, trimming_level)
        trimmed_mean_2 = stats.trim_mean(column_2, trimming_level)
        winsorized_standard_deviation_1 = np.sqrt(utils.winsorized_variance(column_1))
        winsorized_standard_deviation_2 = np.sqrt(utils.winsorized_variance(column_2))

        standardizer = np.sqrt(
            (
                (
                    winsorized_standard_deviation_1**2 * (sample_size_1 - 1)
                    + winsorized_standard_deviation_2**2 * (sample_size_2 - 1)
                )
                / (sample_size - 2)
            )
        )
        akp_effect_size = (
            correction
            * ((trimmed_mean_1 - trimmed_mean_2) - population_difference)
            / standardizer
        )

        q = sample_size_1 / sample_size
        kms = (
            (trimmed_mean_1 - trimmed_mean_2)
            / (
                (1 - q) * winsorized_standard_deviation_1**2
                + q * winsorized_standard_deviation_2**2
            )
            / (q * (1 - q))
        )

        bootstrap_samples_x = []
        bootstrap_samples_y = []
        for _ in range(reps):
            sample_1_bootstrap = np.random.choice(column_1, len(column_1), replace=True)
            sample_2_bootstrap = np.random.choice(column_2, len(column_2), replace=True)
            bootstrap_samples_x.append(sample_1_bootstrap)
            bootstrap_samples_y.append(sample_2_bootstrap)

        trimmed_means_of_bootstrap_sample_1 = np.array(
            (stats.trim_mean(bootstrap_samples_x, trimming_level, axis=1))
        )
        trimmed_means_of_bootstrap_sample_2 = np.array(
            (stats.trim_mean(bootstrap_samples_y, trimming_level, axis=1))
        )
        winsorized_variances_of_bootstrap_sample_1 = np.array(
            (
                [
                    utils.winsorized_variance(arr, trimming_level)
                    for arr in bootstrap_samples_x
                ]
            )
        )
        winsorized_variances_of_bootstrap_sample_2 = np.array(
            (
                [
                    utils.winsorized_variance(arr, trimming_level)
                    for arr in bootstrap_samples_y
                ]
            )
        )
        standardizer_of_bootstrap = np.array(
            (
                [
                    (x * (sample_size_1 - 1) + y * (sample_size_2 - 1))
                    / (sample_size - 2)
                    for x, y in zip(
                        winsorized_variances_of_bootstrap_sample_1,
                        winsorized_variances_of_bootstrap_sample_2,
                    )
                ]
            )
        )
        sample_difference_of_bootstrap = np.array(
            (
                [
                    (x - y)
                    for x, y in zip(
                        trimmed_means_of_bootstrap_sample_1,
                        trimmed_means_of_bootstrap_sample_2,
                    )
                ]
            )
        )
        akp_effect_size_bootstrap = [
            correction * (x - population_difference) / np.sqrt(y)
            for x, y in zip(sample_difference_of_bootstrap, standardizer_of_bootstrap)
        ]

        lower_ci_akp_boot = np.percentile(
            akp_effect_size_bootstrap,
            ((1 - confidence_level) - ((1 - confidence_level) / 2)) * 100,
        )
        upper_ci_akp_boot = np.percentile(
            akp_effect_size_bootstrap,
            ((confidence_level) + ((1 - confidence_level) / 2)) * 100,
        )

        sort_values = np.concatenate((column_1, column_2))
        variance_between_trimmed_means = (
            np.std(np.array([trimmed_mean_1, trimmed_mean_2]), ddof=1)
        ) ** 2
        winsorized_variance_value = utils.winsorized_variance(
            sort_values, trimming_level
        )
        explained_variance = variance_between_trimmed_means / (
            winsorized_variance_value / correction**2
        )
        explanatory_power_effect_size = np.sqrt(explained_variance)

        bootstrap_samples_x = []
        bootstrap_samples_y = []
        for _ in range(reps):
            sample_1_bootstrap = np.random.choice(column_1, len(column_1), replace=True)
            sample_2_bootstrap = np.random.choice(column_2, len(column_2), replace=True)
            bootstrap_samples_x.append(sample_1_bootstrap)
            bootstrap_samples_y.append(sample_2_bootstrap)

        concatenated_samples = [
            np.concatenate((x, y))
            for x, y in zip(bootstrap_samples_x, bootstrap_samples_y)
        ]
        trimmed_means_of_bootstrap_sample_1 = np.array(
            (stats.trim_mean(bootstrap_samples_x, trimming_level, axis=1))
        )
        trimmed_means_of_bootstrap_sample_2 = np.array(
            (stats.trim_mean(bootstrap_samples_y, trimming_level, axis=1))
        )
        variance_between_trimmed_means_bootstrap = [
            (np.std(np.array([x, y]), ddof=1)) ** 2
            for x, y in zip(
                trimmed_means_of_bootstrap_sample_1, trimmed_means_of_bootstrap_sample_2
            )
        ]
        winsorized_variances_bootstrap = [
            utils.winsorized_variance(arr, trimming_level)
            for arr in concatenated_samples
        ]
        explained_variance_bootstrapping = np.array(
            variance_between_trimmed_means_bootstrap
            / (winsorized_variances_bootstrap / correction**2)
        )
        explanatory_power_effect_size_bootstrap = [
            array**0.5 for array in explained_variance_bootstrapping
        ]
        lower_ci_e_pow_boot = np.percentile(
            explanatory_power_effect_size_bootstrap,
            ((1 - confidence_level) - ((1 - confidence_level) / 2)) * 100,
        )
        upper_ci_e_pow_boot = np.percentile(
            explanatory_power_effect_size_bootstrap,
            ((confidence_level) + ((1 - confidence_level) / 2)) * 100,
        )

        h1 = sample_size_1 - 2 * np.floor(trimming_level * sample_size_1)
        h2 = sample_size_2 - 2 * np.floor(trimming_level * sample_size_2)
        difference_trimmed_means = trimmed_mean_1 - trimmed_mean_2
        lower_bound = stats.norm.ppf(trimming_level)
        upper_bound = stats.norm.ppf(1 - trimming_level)
        area = utils.area_under_function(
            utils.density, float(lower_bound), float(upper_bound)
        )
        tail_correction = 2 * (lower_bound**2) * trimming_level
        correction = area + tail_correction

        q1 = (sample_size_1 - 1) * winsorized_standard_deviation_1**2 / (h1 * (h1 - 1))
        q2 = (sample_size_2 - 1) * winsorized_standard_deviation_2**2 / (h2 * (h2 - 1))
        df = (q1 + q2) ** 2 / ((q1**2 / (h1 - 1)) + (q2**2 / (h2 - 1)))
        yuen_standard_error = np.sqrt(q1 + q2)
        yuen_statistic = np.abs(difference_trimmed_means / yuen_standard_error)
        yuen_p_value = 2 * (1 - stats.t.cdf(abs(yuen_statistic), df))

        sample1: res.Sample = res.Sample(
            mean=round(trimmed_mean_1, 4),
            standard_deviation=round(winsorized_standard_deviation_1, 4),
            size=round(sample_size_1),
        )
        sample2: res.Sample = res.Sample(
            mean=round(trimmed_mean_2, 4),
            standard_deviation=round(winsorized_standard_deviation_2, 4),
            size=round(sample_size_2),
        )
        inferential: res.InferentialStatistics = res.InferentialStatistics(
            p_value=float(np.around(yuen_p_value, 4)),
            score=round(yuen_statistic, 4),
        )
        inferential.degrees_of_freedom = round(df, 4)
        inferential.means_difference = round(difference_trimmed_means, 4)
        inferential.standard_error = round(yuen_standard_error, 4)

        robust_akp: es.RobustAKP = es.RobustAKP(
            value=round(float(akp_effect_size), 4),
            ci_lower=float(lower_ci_akp_boot),
            ci_upper=float(upper_ci_akp_boot),
            standard_error=round(yuen_standard_error, 4),
        )

        robust_exp: es.RobustExplanatory = es.RobustExplanatory(
            value=round(min(float(explanatory_power_effect_size), 1.0), 4),
            ci_lower=round(float(lower_ci_e_pow_boot), 4),
            ci_upper=round(min(float(upper_ci_e_pow_boot), 1.0), 4),
            standard_error=round(yuen_standard_error, 4),
        )

        results = TwoIndependentRobustResults()
        results.sample1 = sample1
        results.sample2 = sample2
        results.inferential = inferential
        results.robust_akp = robust_akp
        results.robust_exp = robust_exp

        return results



"""
Two Independent Means with Unequal Variances Calculator
"""

from dataclasses import dataclass
from typing import Optional
import math
import numpy as np
from scipy import stats
from ...utils import interfaces, res, utils, es, texts


@dataclass
class TwoIndependentUnequalVarResults:
    """
    Results for the Two Independent Means with Unequal Variances test.
    """

    inferential: Optional[res.InferentialStatistics] = None
    aoki_epsilon: Optional[es.AokiEpsilon] = None
    aoki_epsilon_unbiased: Optional[es.AokiEpsilonUnbiased] = None
    sample1: Optional[res.Sample] = None
    sample2: Optional[res.Sample] = None


class TwoIndependentUnequalVarTests(interfaces.AbstractTest):
    """
    Two Independent Means with Unequal Variances test.
    This test evaluates the difference between two independent means when the variances are unequal.
    It calculates Welch's t-test, degrees of freedom, p-value, and Aoki's Epsilon effect sizes.
    This class provides methods to compute the test from parameters, data, or a score.

        Attributes:
                None

        Methods:
                from_score: Computes the test from a score (not implemented).
                from_parameters: Computes the test from sample means, standard deviations, sizes, and population mean difference.
                from_data: Computes the test from two independent samples.
    """

    @utils.not_implemented(
        interfaces.MethodType.FROM_SCORE, "TwoIndependentUnequalVarTests"
    )
    @staticmethod
    def from_score() -> None:
        """
        A static method to create results from a score.
        This method is not implemented and will raise a NotImplementedError.
        Raises:
                NotImplementedError: This method is not implemented for TwoIndependentUnequalVar.
        """
        pass

    @staticmethod
    def from_parameters(
        sample_mean_1: float,
        sample_mean_2: float,
        sample_sd_1: float,
        sample_sd_2: float,
        sample_size_1: int,
        sample_size_2: int,
        population_mean_diff: float = 0,
        confidence_level: float = 0.95,
    ) -> TwoIndependentUnequalVarResults:
        """
        Computes the test from parameters.

        Arguments:
            sample_mean_1 (float): The mean of the first sample.
            sample_mean_2 (float): The mean of the second sample.
            sample_sd_1 (float): The standard deviation of the first sample.
            sample_sd_2 (float): The standard deviation of the second sample.
            sample_size_1 (int): The size of the first sample.
            sample_size_2 (int): The size of the second sample.
            population_mean_diff (float): The hypothesized difference in population means. Default is 0.
            confidence_level (float): The confidence level for the confidence intervals. Default is 0.95.

        Returns:
            TwoIndependentUnequalVarResults: The results of the test.
        """
        sample_size = sample_size_1 + sample_size_2
        mean_difference = sample_mean_1 - sample_mean_2
        variance_1 = sample_sd_1**2
        variance_2 = sample_sd_2**2
        standard_error_welch_t = np.sqrt(
            variance_1 / sample_size_1 + (variance_2 / sample_size_2)
        )
        welchs_t = (
            sample_mean_1 - sample_mean_2 - population_mean_diff
        ) / standard_error_welch_t
        df = (variance_1 / sample_size_1 + variance_2 / sample_size_2) ** 2 / (
            (variance_1 / sample_size_1) ** 2 / (sample_size_1 - 1)
            + (variance_2 / sample_size_2) ** 2 / (sample_size_2 - 1)
        )
        p_value = min(float(stats.t.sf((abs(welchs_t)), df) * 2), 0.99999)

        harmonic_n = sample_size_1 * sample_size_2 / (sample_size_1 + sample_size_2)
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        epsilon = welchs_t / np.sqrt(harmonic_n)
        epsilon_unbiased = epsilon * correction
        standard_error_epsilon_biased = (
            df / (df - 2) * (1 / harmonic_n + epsilon_unbiased**2)
            - epsilon_unbiased**2 / correction**2
        )
        standard_error_epsilon_unbiased = standard_error_epsilon_biased * correction**2
        standardizer_biased = np.sqrt(
            (variance_1 * sample_size_2 + variance_2 * sample_size_1)
            / (sample_size_1 + sample_size_2)
        )
        standardizer_unbiased = standardizer_biased / correction

        lower_ncp, upper_ncp = utils.pivotal_ci_t(
            welchs_t, df, sample_size, confidence_level
        )

        epsilon_biased_lower_ci = (lower_ncp * (np.sqrt(sample_size))) / np.sqrt(
            harmonic_n
        )
        epsilon_biased_upper_ci = (upper_ncp * (np.sqrt(sample_size))) / np.sqrt(
            harmonic_n
        )
        epsilon_unbiased_lower_ci = (
            (lower_ncp * (np.sqrt(sample_size))) / np.sqrt(harmonic_n) * correction
        )
        epsilon_unbiased_upper_ci = (
            (upper_ncp * (np.sqrt(sample_size))) / np.sqrt(harmonic_n) * correction
        )

        inferential: res.InferentialStatistics = res.InferentialStatistics(
            score=round(welchs_t, 4),
            p_value=round(p_value, 4),
        )
        inferential.degrees_of_freedom = round(df, 4)
        inferential.standard_error = round(standard_error_welch_t, 4)
        inferential.means_difference = round(mean_difference, 4)

        aoki_epsilon: es.AokiEpsilon = es.AokiEpsilon(
            value=round(epsilon, 4),
            standard_error=round(standard_error_epsilon_biased, 4),
            ci_lower=round(epsilon_biased_lower_ci, 4),
            ci_upper=round(epsilon_biased_upper_ci, 4),
        )
        aoki_epsilon.standardizer = round(standardizer_biased, 4)

        aoki_epsilon_unbiased: es.AokiEpsilonUnbiased = es.AokiEpsilonUnbiased(
            value=round(epsilon_unbiased, 4),
            standard_error=round(standard_error_epsilon_unbiased, 4),
            ci_lower=round(epsilon_unbiased_lower_ci, 4),
            ci_upper=round(epsilon_unbiased_upper_ci, 4),
        )
        aoki_epsilon_unbiased.standardizer = round(standardizer_unbiased, 4)

        results = TwoIndependentUnequalVarResults()
        results.inferential = inferential
        results.aoki_epsilon = aoki_epsilon
        results.aoki_epsilon_unbiased = aoki_epsilon_unbiased

        return results

    @staticmethod
    def from_data(
        columns: list[list], population_mean_diff: float, confidence_level: float = 0.95
    ) -> TwoIndependentUnequalVarResults:
        """
        Computes the test from data.

        Arguments:
            columns (list[list]): A list of two columns of data.
            population_mean_diff (float): The hypothesized difference in population means. Default is 0.
            confidence_level (float): The confidence level for the confidence intervals. Default is 0.95.

        Returns:
            TwoIndependentUnequalVarResults: The results of the test.
        """
        if len(columns) != 2:
            raise ValueError(texts.Errors.columns_must_be_two)

        column_1 = columns[0]
        column_2 = columns[1]

        sample_mean_1 = np.mean(column_1)
        sample_mean_2 = np.mean(column_2)
        sample_sd_1 = np.std(column_1, ddof=1)
        sample_sd_2 = np.std(column_2, ddof=1)
        sample_size = len(column_1)
        sample_size_1 = len(column_1)
        sample_size_2 = len(column_2)

        sample_size = sample_size_1 + sample_size_2
        mean_difference = sample_mean_1 - sample_mean_2
        variance_1 = sample_sd_1**2
        variance_2 = sample_sd_2**2

        standard_error_welch_t = np.sqrt(
            variance_1 / sample_size_1 + (variance_2 / sample_size_2)
        )
        welchs_t = (
            sample_mean_1 - sample_mean_2 - population_mean_diff
        ) / standard_error_welch_t
        df = (variance_1 / sample_size_1 + variance_2 / sample_size_2) ** 2 / (
            (variance_1 / sample_size_1) ** 2 / (sample_size_1 - 1)
            + (variance_2 / sample_size_2) ** 2 / (sample_size_2 - 1)
        )
        p_value = min(float(stats.t.sf((abs(welchs_t)), df) * 2), 0.99999)

        harmonic_n = sample_size_1 * sample_size_2 / (sample_size_1 + sample_size_2)
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )

        epsilon = welchs_t / np.sqrt(harmonic_n)
        epsilon_unbiased = epsilon * correction

        standard_error_epsilon_biased = (
            df / (df - 2) * (1 / harmonic_n + epsilon_unbiased**2)
            - epsilon_unbiased**2 / correction**2
        )
        standard_error_epsilon_unbiased = standard_error_epsilon_biased * correction**2

        standardizer_biased = np.sqrt(
            (variance_1 * sample_size_2 + variance_2 * sample_size_1)
            / (sample_size_1 + sample_size_2)
        )
        standardizer_unbiased = standardizer_biased / correction

        lower_ncp, upper_ncp = utils.pivotal_ci_t(
            welchs_t, float(df), sample_size, confidence_level
        )

        epsilon_biased_lower_ci = (lower_ncp * (np.sqrt(sample_size))) / np.sqrt(
            harmonic_n
        )
        epsilon_biased_upper_ci = (upper_ncp * (np.sqrt(sample_size))) / np.sqrt(
            harmonic_n
        )
        epsilon_unbiased_lower_ci = (
            (lower_ncp * (np.sqrt(sample_size))) / np.sqrt(harmonic_n) * correction
        )
        epsilon_unbiased_upper_ci = (
            (upper_ncp * (np.sqrt(sample_size))) / np.sqrt(harmonic_n) * correction
        )

        sample1: res.Sample = res.Sample(
            mean=float(sample_mean_1),
            standard_deviation=float(sample_sd_1),
            size=sample_size_1,
        )

        sample2: res.Sample = res.Sample(
            mean=float(sample_mean_2),
            standard_deviation=float(sample_sd_2),
            size=sample_size_2,
        )

        inferential: res.InferentialStatistics = res.InferentialStatistics(
            score=round(welchs_t, 4),
            p_value=round(p_value, 4),
        )
        inferential.degrees_of_freedom = round(float(df), 4)
        inferential.means_difference = round(float(mean_difference), 4)

        aoki_epsilon: es.AokiEpsilon = es.AokiEpsilon(
            value=round(epsilon, 4),
            standard_error=round(standard_error_epsilon_biased, 4),
            ci_lower=round(epsilon_biased_lower_ci, 4),
            ci_upper=round(epsilon_biased_upper_ci, 4),
        )
        aoki_epsilon.standardizer = round(standardizer_biased, 4)

        aoki_epsilon_unbiased: es.AokiEpsilonUnbiased = es.AokiEpsilonUnbiased(
            value=round(epsilon_unbiased, 4),
            standard_error=round(standard_error_epsilon_unbiased, 4),
            ci_lower=round(epsilon_unbiased_lower_ci, 4),
            ci_upper=round(epsilon_unbiased_upper_ci, 4),
        )
        aoki_epsilon_unbiased.standardizer = round(standardizer_unbiased, 4)

        results = TwoIndependentUnequalVarResults()
        results.inferential = inferential
        results.sample1 = sample1
        results.sample2 = sample2
        results.aoki_epsilon = aoki_epsilon
        results.aoki_epsilon_unbiased = aoki_epsilon_unbiased

        return results



"""
Two Independent Z Tests for Two Independent Means
"""

from dataclasses import dataclass
from typing import Optional
import numpy as np
from scipy import stats
from ...utils import interfaces, res, utils, es, texts


@dataclass
class TwoIndependentZResults:
    """
    Results for Two Independent Z Tests.
    """

    inferential: Optional[res.InferentialStatistics] = None
    cohens_d: Optional[es.CohenD] = None
    sample1: Optional[res.Sample] = None
    sample2: Optional[res.Sample] = None


class TwoIndependentZTests(interfaces.AbstractTest):
    """
    Two Independent Z Tests for Two Independent Means.
    This class provides methods to calculate Z scores, Cohen's d, and confidence intervals
    for two independent samples.

    Attributes
    ----------
    None

    Methods
    -------
    from_score() - Create an instance from a Z score.
    from_parameters() - Create an instance from sample parameters.
    from_data() - Create an instance from raw data.
    """

    @staticmethod
    def from_score(
        z_score: float,
        sample_size_1: int,
        sample_size_2: int,
        confidence_level: float = 0.95,
    ) -> TwoIndependentZResults:
        """
        Create an instance from a Z score.

        Args
        ----
        z_score: float
                The Z score to use.
        sample_size_1: int
                The sample size of the first group.
        sample_size_2: int
                The sample size of the second group.
        confidence_level: float
                The confidence level for the confidence interval (default is 0.95).

        Returns
        -------
        TwoIndependentZResults
            The results of the two independent Z test.
        """
        total_sample_size = sample_size_1
        mean_sample_size = (sample_size_1 + sample_size_2) / 2
        cohens_d = ((2 * z_score) / np.sqrt(total_sample_size)) * np.sqrt(
            mean_sample_size / ((2 * sample_size_1 * sample_size_2) / total_sample_size)
        )
        p_value = min(float(stats.norm.sf((abs(z_score))) * 2), 0.99999)
        ci_lower, ci_upper, standard_error_es = utils.ci_from_cohens_d_two_samples(
            cohens_d, sample_size_1, sample_size_2, confidence_level
        )

        inferential: res.InferentialStatistics = res.InferentialStatistics(
            score=round(z_score, 4),
            p_value=round(p_value, 4),
        )

        cohens_d_es: es.CohenD = es.CohenD(
            value=round(cohens_d, 4),
            ci_lower=round(ci_lower, 4),
            ci_upper=round(ci_upper, 4),
            standard_error=round(standard_error_es, 4),
        )

        results = TwoIndependentZResults()
        results.inferential = inferential
        results.cohens_d = cohens_d_es

        return results

    @staticmethod
    def from_parameters(
        sample_mean_1: float,
        sample_mean_2: float,
        population_sd_1: float,
        population_sd_2: float,
        sample_size_1: int,
        sample_size_2: int,
        population_diff: float,
        confidence_level: float = 0.95,
    ) -> TwoIndependentZResults:
        """
        Create an instance from a parameters.

        Args
        ----
        sample_mean_1: float
                The mean of the first sample.
        sample_mean_2: float
                The mean of the second sample.
        population_sd_1: float
                The population standard deviation of the first sample.
        population_sd_2: float
                The population standard deviation of the second sample.
        sample_size_1: int
                The sample size of the first group.
        sample_size_2: int
                The sample size of the second group.
        population_diff: float
                The population difference to test.
        confidence_level: float
                The confidence level for the confidence interval (default is 0.95).

        Returns
        -------
        TwoIndependentZResults
            The results of the two independent Z test.
        """
        var1 = population_sd_1**2
        var2 = population_sd_2**2
        standard_error_mean_difference_population = np.sqrt(
            ((var1 / sample_size_1) + (var2 / sample_size_2))
        )
        sd_difference_pop = np.sqrt((var1 + var2) / 2)
        z_score = (
            population_diff - (sample_mean_1 - sample_mean_2)
        ) / standard_error_mean_difference_population
        cohens_d = abs(
            (population_diff - (sample_mean_1 - sample_mean_2)) / sd_difference_pop
        )
        p_value = min(float(stats.norm.sf((abs(z_score))) * 2), 0.99999)
        ci_lower, ci_upper, standard_error_es = utils.ci_from_cohens_d_two_samples(
            cohens_d, sample_size_1, sample_size_2, confidence_level
        )

        sample1: res.Sample = res.Sample(
            mean=round(sample_mean_1, 4),
            size=round(sample_size_1, 4),
            standard_deviation=round(population_sd_1, 4),
        )

        sample1.diff_mean = round(population_diff, 4)
        sample1.diff_sd = round(sd_difference_pop, 4)

        sample2: res.Sample = res.Sample(
            mean=round(sample_mean_2, 4),
            size=round(sample_size_2, 4),
            standard_deviation=round(population_sd_2, 4),
        )

        sample2.diff_mean = round(population_diff, 4)
        sample2.diff_sd = round(sd_difference_pop, 4)

        inferential: res.InferentialStatistics = res.InferentialStatistics(
            score=round(z_score, 4),
            p_value=round(p_value, 4),
        )
        inferential.standard_error = round(standard_error_mean_difference_population, 4)

        cohens_d_es: es.CohenD = es.CohenD(
            value=round(cohens_d, 4),
            ci_lower=round(ci_lower, 4),
            ci_upper=round(ci_upper, 4),
            standard_error=round(standard_error_es, 4),
        )

        results = TwoIndependentZResults()
        results.sample1 = sample1
        results.sample2 = sample2
        results.inferential = inferential
        results.cohens_d = cohens_d_es

        return results

    @staticmethod
    def from_data(
        columns: list[list],
        population_diff: float,
        population_sd_1: float,
        population_sd_2: float,
        confidence_level: float = 0.95,
    ) -> TwoIndependentZResults:
        """
        Create an instance from a parameters.

        Args
        ----
        columns: list[list]
                The data columns to use for the test.
        population_diff: float
                The population difference to test.
        population_sd_1: float
                The population standard deviation of the first sample.
        population_sd_2: float
                The population standard deviation of the second sample..
        confidence_level: float
                The confidence level for the confidence interval (default is 0.95).

        Returns
        -------
        TwoIndependentZResults
            The results of the two independent Z test.

        Raises
        ------
        ValueError
            If the columns list does not contain exactly two columns.
        """
        if len(columns) != 2:
            raise ValueError(texts.Errors.columns_must_be_two)

        column_1 = columns[0]
        column_2 = columns[1]

        sample_mean_1 = np.mean(column_1)
        sample_mean_2 = np.mean(column_2)
        sample_std_1 = np.std(column_1, ddof=1)
        sample_std_2 = np.std(column_2, ddof=1)
        var1 = population_sd_1**2
        var2 = population_sd_2**2
        sample_size_1 = len(column_1)
        sample_size_2 = len(column_2)
        standard_error_mean_difference_population = np.sqrt(
            ((var1 / sample_size_1) + (var2 / sample_size_2))
        )
        sd_difference_pop = np.sqrt((var1 + var2) / 2)
        z_score = (
            population_diff - (sample_mean_1 - sample_mean_2)
        ) / standard_error_mean_difference_population
        cohens_d = abs(
            (population_diff - (sample_mean_1 - sample_mean_2)) / sd_difference_pop
        )
        p_value = min(float(stats.norm.sf((abs(z_score))) * 2), 0.99999)
        ci_lower, ci_upper, standard_error_es = utils.ci_from_cohens_d_two_samples(
            cohens_d, sample_size_1, sample_size_2, confidence_level
        )

        sample1: res.Sample = res.Sample(
            mean=round(float(sample_mean_1), 4),
            size=round(sample_size_1, 4),
            standard_deviation=round(float(sample_std_1), 4),
        )
        sample1.population_sd_diff = round(population_sd_1, 4)
        sample1.diff_sd = round(sd_difference_pop, 4)

        sample2: res.Sample = res.Sample(
            mean=round(float(sample_mean_2), 4),
            size=round(sample_size_2, 4),
            standard_deviation=round(float(sample_std_2), 4),
        )
        sample2.population_sd_diff = round(population_sd_2, 4)
        sample2.diff_sd = round(sd_difference_pop, 4)

        inferential: res.InferentialStatistics = res.InferentialStatistics(
            score=round(z_score, 4),
            p_value=round(p_value, 4),
        )
        inferential.standard_error = round(standard_error_mean_difference_population, 4)

        cohens_d_es: es.CohenD = es.CohenD(
            value=round(cohens_d, 4),
            ci_lower=round(ci_lower, 4),
            ci_upper=round(ci_upper, 4),
            standard_error=round(standard_error_es, 4),
        )

        results = TwoIndependentZResults()
        results.sample1 = sample1
        results.sample2 = sample2
        results.inferential = inferential
        results.cohens_d = cohens_d_es

        return results



"""
This module implements the Two Paired T-test for comparing two related samples.
It provides methods to calculate the test statistics, confidence intervals,
and effect sizes such as Cohen's d, Hedges' g, and their variants.
"""

import math
from dataclasses import dataclass
from typing import Optional
import numpy as np
from scipy import stats
from ...utils import interfaces, res, utils, es


@dataclass
class TwoPairedTResults:
    """
    A class to store the results of a Two Paired T-test.
    """

    cohens_d: Optional[es.CohenD] = None
    hedge_g: Optional[es.HedgesG] = None
    cohens_dav: Optional[es.CohensDav] = None
    hedge_gav: Optional[es.HedgesGav] = None
    cohens_drm: Optional[es.CohensDrm] = None
    hedge_grm: Optional[es.HedgesGrm] = None
    ratio_of_means: Optional[es.RatioOfMeans] = None
    inferential: Optional[res.InferentialStatistics] = None
    sample1: Optional[res.Sample] = None
    sample2: Optional[res.Sample] = None


class TwoPairedTTests(interfaces.AbstractTest):
    """
    A class to perform Two Paired T-tests and calculate effect sizes.
    This class provides methods to calculate Cohen's d, Hedges' g, and their variants
    based on the t-score, sample size, and confidence level.
    It also provides methods to calculate confidence intervals and standard errors
    for the effect sizes.
    It can be used to analyze paired samples and compute the necessary statistics
    for hypothesis testing in a two-sample context.

    Attributes:
        None
    Methods:
        from_score -> TwoPairedTResults:
            Calculates effect sizes and confidence intervals from a t-score.
        from_parameters -> TwoPairedTResults:
            Calculates effect sizes and confidence intervals from sample parameters.
        from_data -> TwoPairedTResults:
            Calculates effect sizes and confidence intervals from data columns.
        calculate_central_ci -> tuple:
            Calculates central confidence intervals and standard errors for effect sizes.
        calculate_se_pooled -> tuple:
            Calculates pooled standard errors and confidence intervals for effect sizes.
        pivotal_ci_t -> tuple:
            Calculates pivotal confidence intervals for effect sizes based on t-distribution.
        ci_t_prime -> tuple:
            Calculates t-prime confidence intervals for effect sizes.
        ci_adjusted_lambda_prime -> tuple:
            Calculates adjusted lambda prime confidence intervals for effect sizes.
        ci_mag -> tuple:
            Calculates magnitude confidence intervals for effect sizes.
        ci_morris -> tuple:
            Calculates Morris confidence intervals for effect sizes.

    """

    @staticmethod
    def from_score(
        t_score: float, sample_size: int, confidence_level: float
    ) -> TwoPairedTResults:
        """
        Calculates effect sizes and confidence intervals from a t-score.
        """

        df = int(sample_size - 1)
        p_value = min(float(stats.t.sf((abs(t_score)), df) * 2), 0.99999)

        cohens_dz_value = t_score / np.sqrt(sample_size)
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_gz_value = correction * cohens_dz_value

        (
            ci_lower_cohens_dz_central,
            ci_upper_cohens_dz_central,
            standard_error_cohens_dz_true,
            standard_error_cohens_dz_morris,
            standard_error_cohens_dz_hedges,
            standard_error_cohens_dz_hedges_olkin,
            standard_error_cohens_dz_mle,
            standard_error_cohens_dz_largen,
            standard_error_cohens_dz_small_n,
        ) = utils.central_ci_paired(cohens_dz_value, sample_size, confidence_level)
        (
            ci_lower_hedges_gz_central,
            ci_upper_hedges_gz_central,
            standard_error_hedges_gz_true,
            standard_error_hedges_gz_morris,
            standard_error_hedges_gz_hedges,
            standard_error_hedges_gz_hedges_olkin,
            standard_error_hedges_gz_mle,
            standard_error_hedges_gz_largen,
            standard_error_hedges_gz_small_n,
        ) = utils.central_ci_paired(hedges_gz_value, sample_size, confidence_level)

        ci_lower_cohens_dz_pivotal, ci_upper_cohens_dz_pivotal = utils.pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_hedges_gz_pivotal, ci_upper_hedges_gz_pivotal = utils.pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )

        cohens_dz = es.CohenD(
            value=cohens_dz_value,
            ci_lower=ci_lower_cohens_dz_central,
            ci_upper=ci_upper_cohens_dz_central,
            standard_error=standard_error_cohens_dz_true,
        )

        hedges_gz = es.HedgesG(
            value=hedges_gz_value,
            ci_lower=ci_lower_hedges_gz_central,
            ci_upper=ci_upper_hedges_gz_central,
            standard_error=standard_error_hedges_gz_true,
        )

        cohens_dz.standardizer = (2 * cohens_dz.value) / t_score
        hedges_gz.standardizer = cohens_dz.standardizer / correction

        cohens_dz.update_pivotal_ci(
            round(ci_lower_cohens_dz_pivotal, 4), round(ci_upper_cohens_dz_pivotal, 4)
        )
        hedges_gz.update_pivotal_ci(
            round(ci_lower_hedges_gz_pivotal * correction, 4),
            round(ci_upper_hedges_gz_pivotal * correction, 4),
        )

        inferential = res.InferentialStatistics(
            score=round(t_score, 4),
            p_value=round(p_value, 4),
        )
        inferential.degrees_of_freedom = round(df, 4)

        results = TwoPairedTResults()
        results.cohens_d = cohens_dz
        results.hedge_g = hedges_gz
        results.inferential = inferential

        return results

    @staticmethod
    def from_parameters(
        sample_mean_1: float,
        sample_mean_2: float,
        sample_sd_1: float,
        sample_sd_2: float,
        sample_size: int,
        correlation: float,
        confidence_level: float,
        population_mean_diff: float = 0,
    ) -> TwoPairedTResults:
        """
        Calculates effect sizes and confidence intervals from sample parameters.
        """

        df = int(sample_size - 1)
        standardizer_dz = np.sqrt(
            sample_sd_1**2
            + sample_sd_2**2
            - 2 * correlation * sample_sd_1 * sample_sd_2
        )
        standardizer_dav = np.sqrt((sample_sd_1**2 + sample_sd_2**2) / 2)
        standardizer_drm = np.sqrt(
            sample_sd_1**2
            + sample_sd_2**2
            - 2 * correlation * sample_sd_1 * sample_sd_2
        ) / np.sqrt(2 * (1 - correlation))
        standard_error = np.sqrt(
            ((sample_sd_1**2 / sample_size) + (sample_sd_2**2 / sample_size))
            - (
                2
                * correlation
                * (sample_sd_1 / np.sqrt(sample_size))
                * (sample_sd_2 / np.sqrt(sample_size))
            )
        )
        t_score = (
            (sample_mean_1 - sample_mean_2) - population_mean_diff
        ) / standard_error
        t_score_av = ((sample_mean_1 - sample_mean_2) - population_mean_diff) / (
            (np.sqrt((sample_sd_1**2 + sample_sd_2**2) / 2)) / np.sqrt(sample_size)
        )

        p_value = min(float(stats.t.sf((abs(t_score)), df) * 2), 0.99999)
        cohens_dz_value = (
            (sample_mean_1 - sample_mean_2) - population_mean_diff
        ) / standardizer_dz
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_gz_value = cohens_dz_value * correction
        cohens_dav_value = (
            (sample_mean_1 - sample_mean_2) - population_mean_diff
        ) / standardizer_dav
        hedges_gav_value = cohens_dav_value * correction
        cohens_drm_value = (
            sample_mean_1 - sample_mean_2 - population_mean_diff
        ) / standardizer_drm
        hedges_grm_value = cohens_drm_value * correction

        (
            ci_lower_cohens_dz_central,
            ci_upper_cohens_dz_central,
            standard_error_cohens_dz_true,
            standard_error_cohens_dz_morris,
            standard_error_cohens_dz_hedges,
            standard_error_cohens_dz_hedges_olkin,
            standard_error_cohens_dz_MLE,
            standard_error_cohens_dz_Largen,
            standard_error_cohens_dz_Small_n,
        ) = utils.central_ci_paired(cohens_dz_value, sample_size, confidence_level)
        (
            ci_lower_hedges_gz_central,
            ci_upper_hedges_gz_central,
            standard_error_hedges_gz_true,
            standard_error_hedges_gz_morris,
            standard_error_hedges_gz_hedges,
            standard_error_hedges_gz_hedges_olkin,
            standard_error_hedges_gz_MLE,
            standard_error_hedges_gz_Largen,
            standard_error_hedges_gz_Small_n,
        ) = utils.central_ci_paired(hedges_gz_value, sample_size, confidence_level)
        (
            ci_lower_cohens_dav_central,
            ci_upper_cohens_dav_central,
            standard_error_cohens_dav_true,
            standard_error_cohens_dav_morris,
            standard_error_cohens_dav_hedges,
            standard_error_cohens_dav_hedges_olkin,
            standard_error_cohens_dav_MLE,
            standard_error_cohens_dav_Largen,
            standard_error_cohens_dav_Small_n,
        ) = utils.central_ci_paired(cohens_dav_value, sample_size, confidence_level)
        (
            ci_lower_hedges_gav_central,
            ci_upper_hedges_gav_central,
            standard_error_hedges_gav_true,
            standard_error_hedges_gav_morris,
            standard_error_hedges_gav_hedges,
            standard_error_hedges_gav_hedges_olkin,
            standard_error_hedges_gav_MLE,
            standard_error_hedges_gav_Largen,
            standard_error_hedges_gav_Small_n,
        ) = utils.central_ci_paired(hedges_gav_value, sample_size, confidence_level)
        (
            ci_lower_cohens_drm_central,
            ci_upper_cohens_drm_central,
            standard_error_cohens_drm_true,
            standard_error_cohens_drm_morris,
            standard_error_cohens_drm_hedges,
            standard_error_cohens_drm_hedges_olkin,
            standard_error_cohens_drm_MLE,
            standard_error_cohens_drm_Largen,
            standard_error_cohens_drm_Small_n,
        ) = utils.central_ci_paired(cohens_drm_value, sample_size, confidence_level)
        (
            ci_lower_hedges_grm_central,
            ci_upper_hedges_grm_central,
            standard_error_hedges_grm_true,
            standard_error_hedges_grm_morris,
            standard_error_hedges_grm_hedges,
            standard_error_hedges_grm_hedges_olkin,
            standard_error_hedges_grm_MLE,
            standard_error_hedges_grm_Largen,
            standard_error_hedges_grm_Small_n,
        ) = utils.central_ci_paired(hedges_gz_value, sample_size, confidence_level)
        (
            ci_lower_cohens_dz_central_pooled,
            ci_upper_cohens_dz_central_pooled,
            standard_error_cohens_dz_true_pooled,
            standard_error_cohens_dz_morris_pooled,
            standard_error_cohens_dz_hedges_pooled,
            standard_error_cohens_dz_hedges_olkin_pooled,
            standard_error_cohens_dz_MLE_pooled,
            standard_error_cohens_dz_Largen_pooled,
            standard_error_cohens_dz_Small_n_pooled,
        ) = utils.calculate_se_pooled(
            cohens_dz_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_hedges_gz_central_pooled,
            ci_upper_hedges_gz_central_pooled,
            standard_error_hedges_gz_true_pooled,
            standard_error_hedges_gz_morris_pooled,
            standard_error_hedges_gz_hedges_pooled,
            standard_error_hedges_gz_hedges_olkin_pooled,
            standard_error_hedges_gz_MLE_pooled,
            standard_error_hedges_gz_Largen_pooled,
            standard_error_hedges_gz_Small_n_pooled,
        ) = utils.calculate_se_pooled(
            hedges_gz_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_cohens_dav_central_pooled,
            ci_upper_cohens_dav_central_pooled,
            standard_error_cohens_dav_true_pooled,
            standard_error_cohens_dav_morris_pooled,
            standard_error_cohens_dav_hedges_pooled,
            standard_error_cohens_dav_hedges_olkin_pooled,
            standard_error_cohens_dav_MLE_pooled,
            standard_error_cohens_dav_Largen_pooled,
            standard_error_cohens_dav_Small_n_pooled,
        ) = utils.calculate_se_pooled(
            cohens_dav_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_hedges_gav_central_pooled,
            ci_upper_hedges_gav_central_pooled,
            standard_error_hedges_gav_true_pooled,
            standard_error_hedges_gav_morris_pooled,
            standard_error_hedges_gav_hedges_pooled,
            standard_error_hedges_gav_hedges_olkin_pooled,
            standard_error_hedges_gav_MLE_pooled,
            standard_error_hedges_gav_Largen_pooled,
            standard_error_hedges_gav_Small_n_pooled,
        ) = utils.calculate_se_pooled(
            hedges_gav_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_cohens_drm_central_pooled,
            ci_upper_cohens_drm_central_pooled,
            standard_error_cohens_drm_true_pooled,
            standard_error_cohens_drm_morris_pooled,
            standard_error_cohens_drm_hedges_pooled,
            standard_error_cohens_drm_hedges_olkin_pooled,
            standard_error_cohens_drm_MLE_pooled,
            standard_error_cohens_drm_Largen_pooled,
            standard_error_cohens_drm_Small_n_pooled,
        ) = utils.calculate_se_pooled(
            cohens_drm_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_hedges_grm_central_pooled,
            ci_upper_hedges_grm_central_pooled,
            standard_error_hedges_grm_true_pooled,
            standard_error_hedges_grm_morris_pooled,
            standard_error_hedges_grm_hedges_pooled,
            standard_error_hedges_grm_hedges_olkin_pooled,
            standard_error_hedges_grm_MLE_pooled,
            standard_error_hedges_grm_Largen_pooled,
            standard_error_hedges_grm_Small_n_pooled,
        ) = utils.calculate_se_pooled(
            hedges_gz_value, sample_size, correlation, confidence_level
        )
        ci_lower_cohens_dz_pivotal, ci_upper_cohens_dz_pivotal = utils.pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_cohens_dav_pivotal, ci_upper_cohens_dav_pivotal = utils.pivotal_ci_t(
            t_score_av, df, sample_size, confidence_level
        )
        ci_lower_cohens_drm_pivotal, ci_upper_cohens_drm_pivotal = utils.pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )

        lower_ci_tprime_dav, upper_ci_tprime_dav = utils.ci_t_prime(
            cohens_dav_value,
            sample_sd_1,
            sample_sd_2,
            sample_size,
            correlation,
            confidence_level,
        )
        lower_ci_tprime_gav, upper_ci_tprime_gav = (
            lower_ci_tprime_dav * correction,
            upper_ci_tprime_dav * correction,
        )
        lower_ci_lambda_prime_dav, upper_ci_lambda_prime_dav = (
            utils.ci_adjusted_lambda_prime(
                cohens_dav_value,
                sample_sd_1,
                sample_sd_2,
                sample_size,
                correlation,
                confidence_level,
            )
        )
        lower_ci_lambda_prime_gav, upper_ci_lambda_prime_gav = (
            lower_ci_lambda_prime_dav * correction,
            upper_ci_lambda_prime_dav * correction,
        )
        lower_ci_mag_dav, upper_ci_mag_dav = utils.ci_mag(
            cohens_dav_value,
            sample_sd_1,
            sample_sd_2,
            sample_size,
            correlation,
            confidence_level,
        )
        lower_ci_mag_gav, upper_ci_mag_gav = (
            lower_ci_mag_dav * correction,
            upper_ci_mag_dav * correction,
        )
        lower_ci_morris_dav, upper_ci_morris_dav = utils.ci_morris(
            cohens_dav_value, sample_size, correlation, confidence_level
        )
        lower_ci_morris_gav, upper_ci_morris_gav = (
            lower_ci_morris_dav * correction,
            upper_ci_morris_dav * correction,
        )

        ratio_of_means = sample_mean_1 / sample_mean_2
        variance_of_means_ratio = (
            sample_sd_1**2 / (sample_mean_1**2)
            + sample_sd_2**2 / (sample_mean_2**2)
            - 2
            * correlation
            * np.sqrt(sample_sd_1**2 * sample_sd_2**2)
            / (sample_mean_1 * sample_mean_2)
        ) / sample_size
        standard_error_of_means_ratio = np.sqrt(variance_of_means_ratio)
        degrees_of_freedom_means_ratio = sample_size - 1
        t_critical_value = stats.t.ppf(
            confidence_level + ((1 - confidence_level) / 2),
            degrees_of_freedom_means_ratio,
        )
        lower_ci_means_ratio = math.exp(
            np.log(ratio_of_means) - t_critical_value * np.sqrt(variance_of_means_ratio)
        )
        upper_ci_means_ratio = math.exp(
            np.log(ratio_of_means) + t_critical_value * np.sqrt(variance_of_means_ratio)
        )

        cohens_dz = es.CohenD(
            value=cohens_dz_value,
            ci_lower=ci_lower_cohens_dz_central,
            ci_upper=ci_upper_cohens_dz_central,
            standard_error=standard_error_cohens_dz_true,
        )
        hedges_gz = es.HedgesG(
            value=hedges_gz_value,
            ci_lower=ci_lower_hedges_gz_central,
            ci_upper=ci_upper_hedges_gz_central,
            standard_error=standard_error_hedges_gz_true,
        )
        cohens_dav = es.CohensDav(
            value=cohens_dav_value,
            ci_lower=ci_lower_cohens_dav_central,
            ci_upper=ci_upper_cohens_dav_central,
            standard_error=standard_error_cohens_dav_true,
        )
        hedges_gav = es.HedgesGav(
            value=hedges_gav_value,
            ci_lower=ci_lower_hedges_gav_central,
            ci_upper=ci_upper_hedges_gav_central,
            standard_error=standard_error_hedges_gav_true,
        )
        cohens_drm = es.CohensDrm(
            value=cohens_drm_value,
            ci_lower=ci_lower_cohens_drm_central,
            ci_upper=ci_upper_cohens_drm_central,
            standard_error=standard_error_cohens_drm_true,
        )
        hedges_grm = es.HedgesGrm(
            value=hedges_grm_value,
            ci_lower=ci_lower_hedges_grm_central,
            ci_upper=ci_upper_hedges_grm_central,
            standard_error=standard_error_hedges_grm_true,
        )
        cohens_dz.standardizer = (2 * cohens_dz.value) / t_score
        if cohens_dz.standardizer:
            hedges_gz.standardizer = cohens_dz.standardizer / correction
        cohens_dav.standardizer = (2 * cohens_dav.value) / t_score_av
        if cohens_dav.standardizer:
            hedges_gav.standardizer = cohens_dav.standardizer / correction
        cohens_drm.standardizer = (2 * cohens_drm.value) / t_score
        if cohens_drm.standardizer:
            hedges_grm.standardizer = cohens_drm.standardizer / correction
        cohens_dz.update_pivotal_ci(
            round(ci_lower_cohens_dz_pivotal, 4), round(ci_upper_cohens_dz_pivotal, 4)
        )
        hedges_gz.update_pivotal_ci(
            round(ci_lower_cohens_dz_pivotal * correction, 4),
            round(ci_upper_cohens_dz_pivotal * correction, 4),
        )
        cohens_dav.update_pivotal_ci(
            round(ci_lower_cohens_dav_pivotal, 4), round(ci_upper_cohens_dav_pivotal, 4)
        )
        hedges_gav.update_pivotal_ci(
            round(ci_lower_cohens_dav_pivotal * correction, 4),
            round(ci_upper_cohens_dav_pivotal * correction, 4),
        )
        cohens_drm.update_pivotal_ci(
            round(ci_lower_cohens_drm_pivotal, 4), round(ci_upper_cohens_drm_pivotal, 4)
        )
        hedges_grm.update_pivotal_ci(
            round(ci_lower_cohens_drm_pivotal * correction, 4),
            round(ci_upper_cohens_drm_pivotal * correction, 4),
        )

        cohens_dav.update_mag_ci(round(lower_ci_mag_dav, 4), round(upper_ci_mag_dav, 4))
        hedges_gav.update_mag_ci(round(lower_ci_mag_gav, 4), round(upper_ci_mag_gav, 4))
        cohens_dav.update_morris_ci(
            round(lower_ci_morris_dav, 4), round(upper_ci_morris_dav, 4)
        )
        hedges_gav.update_morris_ci(
            round(lower_ci_morris_gav, 4), round(upper_ci_morris_gav, 4)
        )
        cohens_dav.update_t_prime_ci(
            round(lower_ci_tprime_dav, 4), round(upper_ci_tprime_dav, 4)
        )
        hedges_gav.update_t_prime_ci(
            round(lower_ci_tprime_gav, 4), round(upper_ci_tprime_gav, 4)
        )
        cohens_dav.update_lambda_prime_ci(
            round(lower_ci_lambda_prime_dav, 4), round(upper_ci_lambda_prime_dav, 4)
        )
        hedges_gav.update_lambda_prime_ci(
            round(lower_ci_lambda_prime_gav, 4), round(upper_ci_lambda_prime_gav, 4)
        )
        ratio_of_means_effect_size = es.RatioOfMeans(
            value=round(ratio_of_means, 4),
            standard_error=round(standard_error_of_means_ratio, 4),
            ci_lower=round(lower_ci_means_ratio, 4),
            ci_upper=round(upper_ci_means_ratio, 4),
        )
        sample1 = res.Sample(
            mean=round(sample_mean_1, 4),
            standard_deviation=int(sample_sd_1),
            size=sample_size,
        )
        sample2 = res.Sample(
            mean=round(sample_mean_2, 4),
            standard_deviation=int(sample_sd_2),
            size=sample_size,
        )

        sample1.diff_mean = round(sample_mean_1 - sample_mean_2, 4)
        sample2.diff_mean = round(sample_mean_2 - sample_mean_1, 4)

        inferential = res.InferentialStatistics(
            score=round(t_score, 4),
            p_value=round(p_value, 4),
        )
        inferential.degrees_of_freedom = round(df, 4)
        inferential.standard_error = round(standard_error, 4)

        results = TwoPairedTResults()
        results.cohens_d = cohens_dz
        results.hedge_g = hedges_gz
        results.cohens_dav = cohens_dav
        results.hedge_gav = hedges_gav
        results.cohens_drm = cohens_drm
        results.hedge_grm = hedges_grm
        results.sample1 = sample1
        results.sample2 = sample2
        results.inferential = inferential
        results.ratio_of_means = ratio_of_means_effect_size

        return results

    @staticmethod
    def from_data(
        columns: list, population_mean_diff: float, confidence_level: float
    ) -> TwoPairedTResults:
        """
        Calculates effect sizes and confidence intervals from data columns.
        """
        column_1 = columns[0]
        column_2 = columns[1]
        sample_mean_1 = np.mean(column_1)
        sample_mean_2 = np.mean(column_2)
        sample_sd_1 = np.std(column_1, ddof=1)
        sample_sd_2 = np.std(column_2, ddof=1)
        sample_size = len(column_1)
        correlation_matrix = np.corrcoef(column_1, column_2)
        correlation = correlation_matrix[0, 1]

        df = int(sample_size - 1)
        standardizer_dz = np.sqrt(
            sample_sd_1**2
            + sample_sd_2**2
            - 2 * correlation * sample_sd_1 * sample_sd_2
        )
        standardizer_dav = np.sqrt((sample_sd_1**2 + sample_sd_2**2) / 2)
        standardizer_drm = np.sqrt(
            sample_sd_1**2
            + sample_sd_2**2
            - 2 * correlation * sample_sd_1 * sample_sd_2
        ) / np.sqrt(2 * (1 - correlation))
        standard_error = np.sqrt(
            ((sample_sd_1**2 / sample_size) + (sample_sd_2**2 / sample_size))
            - (
                2
                * correlation
                * (sample_sd_1 / np.sqrt(sample_size))
                * (sample_sd_2 / np.sqrt(sample_size))
            )
        )
        t_score = (
            (sample_mean_1 - sample_mean_2) - population_mean_diff
        ) / standard_error
        t_score_av = ((sample_mean_1 - sample_mean_2) - population_mean_diff) / (
            (np.sqrt((sample_sd_1**2 + sample_sd_2**2) / 2)) / np.sqrt(sample_size)
        )
        p_value = min(float(stats.t.sf((abs(t_score)), df) * 2), 0.99999)
        cohens_dz_value = (
            (sample_mean_1 - sample_mean_2) - population_mean_diff
        ) / standardizer_dz
        correction = math.exp(
            math.lgamma(df / 2)
            - math.log(math.sqrt(df / 2))
            - math.lgamma((df - 1) / 2)
        )
        hedges_gz_value = cohens_dz_value * correction
        cohens_dav_value = (
            (sample_mean_1 - sample_mean_2) - population_mean_diff
        ) / standardizer_dav
        hedges_gav_value = cohens_dav_value * correction
        cohens_drm_value = (
            sample_mean_1 - sample_mean_2 - population_mean_diff
        ) / standardizer_drm
        hedges_grm_value = cohens_drm_value * correction

        (
            ci_lower_cohens_dz_central,
            ci_upper_cohens_dz_central,
            standard_error_cohens_dz_true,
            standard_error_cohens_dz_morris,
            standard_error_cohens_dz_hedges,
            standard_error_cohens_dz_hedges_olkin,
            standard_error_cohens_dz_mle,
            standard_error_cohens_dz_largen,
            standard_error_cohens_dz_small_n,
        ) = utils.central_ci_paired(cohens_dz_value, sample_size, confidence_level)
        (
            ci_lower_hedges_gz_central,
            ci_upper_hedges_gz_central,
            standard_error_hedges_gz_true,
            standard_error_hedges_gz_morris,
            standard_error_hedges_gz_hedges,
            standard_error_hedges_gz_hedges_olkin,
            standard_error_hedges_gz_mle,
            standard_error_hedges_gz_largen,
            standard_error_hedges_gz_small_n,
        ) = utils.central_ci_paired(hedges_gz_value, sample_size, confidence_level)
        (
            ci_lower_cohens_dav_central,
            ci_upper_cohens_dav_central,
            standard_error_cohens_dav_true,
            standard_error_cohens_dav_morris,
            standard_error_cohens_dav_hedges,
            standard_error_cohens_dav_hedges_olkin,
            standard_error_cohens_dav_mle,
            standard_error_cohens_dav_largen,
            standard_error_cohens_dav_small_n,
        ) = utils.central_ci_paired(cohens_dav_value, sample_size, confidence_level)
        (
            ci_lower_hedges_gav_central,
            ci_upper_hedges_gav_central,
            standard_error_hedges_gav_true,
            standard_error_hedges_gav_morris,
            standard_error_hedges_gav_hedges,
            standard_error_hedges_gav_hedges_olkin,
            standard_error_hedges_gav_mle,
            standard_error_hedges_gav_largen,
            standard_error_hedges_gav_small_n,
        ) = utils.central_ci_paired(hedges_gav_value, sample_size, confidence_level)
        (
            ci_lower_cohens_drm_central,
            ci_upper_cohens_drm_central,
            standard_error_cohens_drm_true,
            standard_error_cohens_drm_morris,
            standard_error_cohens_drm_hedges,
            standard_error_cohens_drm_hedges_olkin,
            standard_error_cohens_drm_mle,
            standard_error_cohens_drm_largen,
            standard_error_cohens_drm_small_n,
        ) = utils.central_ci_paired(cohens_drm_value, sample_size, confidence_level)
        (
            ci_lower_hedges_grm_central,
            ci_upper_hedges_grm_central,
            standard_error_hedges_grm_true,
            standard_error_hedges_grm_morris,
            standard_error_hedges_grm_hedges,
            standard_error_hedges_grm_hedges_olkin,
            standard_error_hedges_grm_mle,
            standard_error_hedges_grm_largen,
            standard_error_hedges_grm_small_n,
        ) = utils.central_ci_paired(hedges_gz_value, sample_size, confidence_level)
        ci_lower_cohens_dz_pivotal, ci_upper_cohens_dz_pivotal = utils.pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        ci_lower_cohens_dav_pivotal, ci_upper_cohens_dav_pivotal = utils.pivotal_ci_t(
            t_score_av, df, sample_size, confidence_level
        )
        ci_lower_cohens_drm_pivotal, ci_upper_cohens_drm_pivotal = utils.pivotal_ci_t(
            t_score, df, sample_size, confidence_level
        )
        (
            ci_lower_cohens_dz_central_pooled,
            ci_upper_cohens_dz_central_pooled,
            standard_error_cohens_dz_true_pooled,
            standard_error_cohens_dz_morris_pooled,
            standard_error_cohens_dz_hedges_pooled,
            standard_error_cohens_dz_hedges_olkin_pooled,
            standard_error_cohens_dz_mle_pooled,
            standard_error_cohens_dz_largen_pooled,
            standard_error_cohens_dz_small_n_pooled,
        ) = utils.calculate_se_pooled(
            cohens_dz_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_hedges_gz_central_pooled,
            ci_upper_hedges_gz_central_pooled,
            standard_error_hedges_gz_true_pooled,
            standard_error_hedges_gz_morris_pooled,
            standard_error_hedges_gz_hedges_pooled,
            standard_error_hedges_gz_hedges_olkin_pooled,
            standard_error_hedges_gz_mle_pooled,
            standard_error_hedges_gz_largen_pooled,
            standard_error_hedges_gz_small_n_pooled,
        ) = utils.calculate_se_pooled(
            hedges_gz_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_cohens_dav_central_pooled,
            ci_upper_cohens_dav_central_pooled,
            standard_error_cohens_dav_true_pooled,
            standard_error_cohens_dav_morris_pooled,
            standard_error_cohens_dav_hedges_pooled,
            standard_error_cohens_dav_hedges_olkin_pooled,
            standard_error_cohens_dav_mle_pooled,
            standard_error_cohens_dav_largen_pooled,
            standard_error_cohens_dav_small_n_pooled,
        ) = utils.calculate_se_pooled(
            cohens_dav_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_hedges_gav_central_pooled,
            ci_upper_hedges_gav_central_pooled,
            standard_error_hedges_gav_true_pooled,
            standard_error_hedges_gav_morris_pooled,
            standard_error_hedges_gav_hedges_pooled,
            standard_error_hedges_gav_hedges_olkin_pooled,
            standard_error_hedges_gav_mle_pooled,
            standard_error_hedges_gav_largen_pooled,
            standard_error_hedges_gav_small_n_pooled,
        ) = utils.calculate_se_pooled(
            hedges_gav_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_cohens_drm_central_pooled,
            ci_upper_cohens_drm_central_pooled,
            standard_error_cohens_drm_true_pooled,
            standard_error_cohens_drm_morris_pooled,
            standard_error_cohens_drm_hedges_pooled,
            standard_error_cohens_drm_hedges_olkin_pooled,
            standard_error_cohens_drm_mle_pooled,
            standard_error_cohens_drm_largen_pooled,
            standard_error_cohens_drm_small_n_pooled,
        ) = utils.calculate_se_pooled(
            cohens_drm_value, sample_size, correlation, confidence_level
        )
        (
            ci_lower_hedges_grm_central_pooled,
            ci_upper_hedges_grm_central_pooled,
            standard_error_hedges_grm_true_pooled,
            standard_error_hedges_grm_morris_pooled,
            standard_error_hedges_grm_hedges_pooled,
            standard_error_hedges_grm_hedges_olkin_pooled,
            standard_error_hedges_grm_mle_pooled,
            standard_error_hedges_grm_largen_pooled,
            standard_error_hedges_grm_small_n_pooled,
        ) = utils.calculate_se_pooled(
            hedges_gz_value, sample_size, correlation, confidence_level
        )

        lower_ci_tprime_dav, upper_ci_tprime_dav = utils.ci_t_prime(
            cohens_dav_value,
            float(sample_sd_1),
            float(sample_sd_2),
            sample_size,
            float(correlation),
            confidence_level,
        )
        lower_ci_tprime_gav, upper_ci_tprime_gav = (
            lower_ci_tprime_dav * correction,
            upper_ci_tprime_dav * correction,
        )
        lower_ci_lambda_prime_dav, upper_ci_lambda_prime_dav = (
            utils.ci_adjusted_lambda_prime(
                cohens_dav_value,
                float(sample_sd_1),
                float(sample_sd_2),
                sample_size,
                float(correlation),
                confidence_level,
            )
        )
        lower_ci_lambda_prime_gav, upper_ci_lambda_prime_gav = (
            lower_ci_lambda_prime_dav * correction,
            upper_ci_lambda_prime_dav * correction,
        )
        lower_ci_mag_dav, upper_ci_mag_dav = utils.ci_mag(
            cohens_dav_value,
            float(sample_sd_1),
            float(sample_sd_2),
            sample_size,
            correlation,
            confidence_level,
        )
        lower_ci_mag_gav, upper_ci_mag_gav = (
            lower_ci_mag_dav * correction,
            upper_ci_mag_dav * correction,
        )
        lower_ci_morris_dav, upper_ci_morris_dav = utils.ci_morris(
            cohens_dav_value, sample_size, correlation, confidence_level
        )
        lower_ci_morris_gav, upper_ci_morris_gav = (
            lower_ci_morris_dav * correction,
            upper_ci_morris_dav * correction,
        )

        ratio_of_means = sample_mean_1 / sample_mean_2
        variance_of_means_ratio = (
            sample_sd_1**2 / (sample_mean_1**2)
            + sample_sd_2**2 / (sample_mean_2**2)
            - 2
            * correlation
            * np.sqrt(sample_sd_1**2 * sample_sd_2**2)
            / (sample_mean_1 * sample_mean_2)
        ) / sample_size
        standard_error_of_means_ratio = np.sqrt(variance_of_means_ratio)
        degrees_of_freedom_means_ratio = sample_size - 1
        t_critical_value = stats.t.ppf(
            confidence_level + ((1 - confidence_level) / 2),
            degrees_of_freedom_means_ratio,
        )
        lower_ci_means_ratio = math.exp(
            np.log(ratio_of_means) - t_critical_value * np.sqrt(variance_of_means_ratio)
        )
        upper_ci_means_ratio = math.exp(
            np.log(ratio_of_means) + t_critical_value * np.sqrt(variance_of_means_ratio)
        )

        cohens_dz = es.CohenD(
            value=cohens_dz_value,
            ci_lower=ci_lower_cohens_dz_central,
            ci_upper=ci_upper_cohens_dz_central,
            standard_error=standard_error_cohens_dz_true,
        )
        hedges_gz = es.HedgesG(
            value=hedges_gz_value,
            ci_lower=ci_lower_hedges_gz_central,
            ci_upper=ci_upper_hedges_gz_central,
            standard_error=standard_error_hedges_gz_true,
        )
        cohens_dav = es.CohensDav(
            value=cohens_dav_value,
            ci_lower=ci_lower_cohens_dav_central,
            ci_upper=ci_upper_cohens_dav_central,
            standard_error=standard_error_cohens_dav_true,
        )
        hedges_gav = es.HedgesGav(
            value=hedges_gav_value,
            ci_lower=ci_lower_hedges_gav_central,
            ci_upper=ci_upper_hedges_gav_central,
            standard_error=standard_error_hedges_gav_true,
        )
        cohens_drm = es.CohensDrm(
            value=cohens_drm_value,
            ci_lower=ci_lower_cohens_drm_central,
            ci_upper=ci_upper_cohens_drm_central,
            standard_error=standard_error_cohens_drm_true,
        )
        hedges_grm = es.HedgesGrm(
            value=hedges_grm_value,
            ci_lower=ci_lower_hedges_grm_central,
            ci_upper=ci_upper_hedges_grm_central,
            standard_error=standard_error_hedges_grm_true,
        )
        cohens_dz.standardizer = standardizer_dz
        if cohens_dz.standardizer:
            hedges_gz.standardizer = cohens_dz.standardizer / correction
        cohens_dav.standardizer = standardizer_dav
        if cohens_dav.standardizer:
            hedges_gav.standardizer = cohens_dav.standardizer / correction
        cohens_drm.standardizer = standardizer_drm
        if cohens_drm.standardizer:
            hedges_grm.standardizer = cohens_drm.standardizer / correction
        cohens_dz.update_pivotal_ci(
            round(ci_lower_cohens_dz_pivotal, 4), round(ci_upper_cohens_dz_pivotal, 4)
        )
        hedges_gz.update_pivotal_ci(
            round(ci_lower_cohens_dz_pivotal * correction, 4),
            round(ci_upper_cohens_dz_pivotal * correction, 4),
        )
        cohens_dav.update_pivotal_ci(
            round(ci_lower_cohens_dav_pivotal, 4), round(ci_upper_cohens_dav_pivotal, 4)
        )
        hedges_gav.update_pivotal_ci(
            round(ci_lower_cohens_dav_pivotal * correction, 4),
            round(ci_upper_cohens_dav_pivotal * correction, 4),
        )
        cohens_drm.update_pivotal_ci(
            round(ci_lower_cohens_drm_pivotal, 4), round(ci_upper_cohens_drm_pivotal, 4)
        )
        hedges_grm.update_pivotal_ci(
            round(ci_lower_cohens_drm_pivotal * correction, 4),
            round(ci_upper_cohens_drm_pivotal * correction, 4),
        )
        cohens_dav.update_mag_ci(round(lower_ci_mag_dav, 4), round(upper_ci_mag_dav, 4))
        hedges_gav.update_mag_ci(round(lower_ci_mag_gav, 4), round(upper_ci_mag_gav, 4))
        cohens_dav.update_morris_ci(
            round(lower_ci_morris_dav, 4), round(upper_ci_morris_dav, 4)
        )
        hedges_gav.update_morris_ci(
            round(lower_ci_morris_gav, 4), round(upper_ci_morris_gav, 4)
        )
        cohens_dav.update_t_prime_ci(
            round(lower_ci_tprime_dav, 4), round(upper_ci_tprime_dav, 4)
        )
        hedges_gav.update_t_prime_ci(
            round(lower_ci_tprime_gav, 4), round(upper_ci_tprime_gav, 4)
        )
        cohens_dav.update_lambda_prime_ci(
            round(lower_ci_lambda_prime_dav, 4), round(upper_ci_lambda_prime_dav, 4)
        )
        hedges_gav.update_lambda_prime_ci(
            round(lower_ci_lambda_prime_gav, 4), round(upper_ci_lambda_prime_gav, 4)
        )
        ratio_of_means_effect_size = es.RatioOfMeans(
            value=round(ratio_of_means, 4),
            standard_error=round(standard_error_of_means_ratio, 4),
            ci_lower=round(lower_ci_means_ratio, 4),
            ci_upper=round(upper_ci_means_ratio, 4),
        )

        sample1 = res.Sample(
            mean=round(sample_mean_1, 4),
            standard_deviation=int(sample_sd_1),
            size=sample_size,
        )
        sample2 = res.Sample(
            mean=round(sample_mean_2, 4),
            standard_deviation=int(sample_sd_2),
            size=sample_size,
        )
        sample1.diff_mean = round(sample_mean_1 - sample_mean_2, 4)
        sample2.diff_mean = round(sample_mean_2 - sample_mean_1, 4)

        inferential = res.InferentialStatistics(
            score=round(t_score, 4),
            p_value=round(p_value, 4),
        )
        inferential.degrees_of_freedom = round(df, 4)
        inferential.standard_error = round(standard_error, 4)

        results = TwoPairedTResults()
        results.cohens_d = cohens_dz
        results.hedge_g = hedges_gz
        results.cohens_dav = cohens_dav
        results.hedge_gav = hedges_gav
        results.cohens_drm = cohens_drm
        results.hedge_grm = hedges_grm
        results.ratio_of_means = ratio_of_means_effect_size
        results.inferential = inferential
        results.sample1 = sample1
        results.sample2 = sample2

        return results



from src.esek.Calculator.OneSampleMean.one_sample_t import (
    OneSampleTResults,
    OneSampleTTest,
)


def test_one_sample_t_from_score():
    data = {
        "t_score": 1.96,
        "sample_size": 30,
        "confidence_level": 0.95,
    }
    try:
        results = OneSampleTTest.from_score(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, OneSampleTResults)


def test_one_sample_t_from_parameters():
    data = {
        "sample_mean": 100,
        "sample_size": 30,
        "sample_sd": 0.5,
        "population_mean": 95,
        "confidence_level": 0.95,
    }
    try:
        results = OneSampleTTest.from_parameters(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, OneSampleTResults)


def test_one_sample_t_from_data():
    assert True, "This test is not implemented yet"



from .two_paired_z import TwoPairedZTests, TwoPairedZResults
from .two_paired_t import TwoPairedTTests, TwoPairedTResults
from .two_paired_aparametric import (
    TwoPairedAparametricTests,
    TwoPairedAparametricResults,
)
from .two_paired_robust import TwoPairedRobustTests, TwoPairedRobustResults
from .two_paired_common_lang import (
    TwoPairedCommonLangTests,
    TwoPairedCommonLangResults,
)

__all__ = [
    "TwoPairedZTests",
    "TwoPairedZResults",
    "TwoPairedTTests",
    "TwoPairedTResults",
    "TwoPairedAparametricTests",
    "TwoPairedAparametricResults",
    "TwoPairedRobustTests",
    "TwoPairedRobustResults",
    "TwoPairedCommonLangTests",
    "TwoPairedCommonLangResults",
]



"""
This module implements the TwoIndependentControlGroupTests class, which provides methods
for conducting statistical tests on two independent control groups.
"""

from dataclasses import dataclass
from typing import Optional
import math
import numpy as np
from scipy import stats
from ...utils import interfaces, res, utils, es, texts


@dataclass
class TwoIndependentControlGroupResults:
    """
    This class holds the results of the two independent control group tests.
    """

    inferential: Optional[res.InferentialStatistics] = None
    glass_delta: Optional[es.GlassDelta] = None
    glass_delta_unbiased: Optional[es.GlassDeltaUnbiased] = None
    experimental_group: Optional[res.Group] = None
    control_group: Optional[res.Group] = None


class TwoIndependentControlGroupTests(interfaces.AbstractTest):
    """
    This class provides methods for conducting statistical tests on two independent control groups.

    Methods:
        - from_score: Not implemented.
        - from_parameters: Creates test results from the given parameters.
        - from_data: Creates test results from the given data.

    """

    @utils.not_implemented(
        interfaces.MethodType.FROM_SCORE, "TwoIndependentControlGroup"
    )
    @staticmethod
    def from_score() -> None:
        """
        A static method to create results from a score.
        This method is not implemented and will raise a NotImplementedError.
        Raises:
                        NotImplementedError: This method is not implemented for TwoIndependentControlGroup.
        """
        pass

    @staticmethod
    def from_parameters(
        sample_mean_experimental: float,
        sample_mean_control: float,
        sample_sd_experimental: float,
        sample_sd_control: float,
        sample_size_experimental: int,
        sample_size_control: int,
        population_mean_diff: float = 0,
        confidence_level: float = 0.95,
    ) -> TwoIndependentControlGroupResults:
        """
        Creates test results from the given parameters.

        Args:
            sample_mean_experimental (float): The mean of the experimental group.
            sample_mean_control (float): The mean of the control group.
            sample_sd_experimental (float): The standard deviation of the experimental group.
            sample_sd_control (float): The standard deviation of the control group.
            sample_size_experimental (int): The sample size of the experimental group.
            sample_size_control (int): The sample size of the control group.
            population_mean_diff (float, optional): The population mean difference. Defaults to 0.
            confidence_level (float, optional): The confidence level for the tests. Defaults to 0.95.

        Returns:
            TwoIndependentControlGroupResults: The results of the tests.
        """
        sample_size = sample_size_experimental + sample_size_control
        df = sample_size - 2
        sample_mean_difference = sample_mean_experimental - sample_mean_control
        standardizer_glass_delta = np.sqrt(
            (
                (((sample_size_experimental - 1) * sample_sd_experimental**2))
                + ((sample_size_control - 1) * sample_sd_control**2)
            )
            / (sample_size - 2)
        )
        standard_error = standardizer_glass_delta * np.sqrt(
            (sample_size_experimental + sample_size_control)
            / (sample_size_experimental * sample_size_control)
        )
        t_score = (sample_mean_difference - population_mean_diff) / standard_error
        t_score_glass = (sample_mean_difference - population_mean_diff) / (
            sample_sd_control / np.sqrt(sample_size_control)
        )
        p_value = min(float(stats.t.sf((abs(t_score)), df) * 2), 0.99999)
        glass_delta = (
            sample_mean_difference - population_mean_diff
        ) / sample_sd_control
        df2 = sample_size_control - 1
        correction = math.exp(
            math.lgamma(df2 / 2)
            - math.log(math.sqrt(df2 / 2))
            - math.lgamma((df2 - 1) / 2)
        )
        unbiased_glass_delta = glass_delta * correction
        standardizer_unbiased_glass_delta = standardizer_glass_delta / correction
        ci_lower_glass_delta_pivotal, ci_upper_glass_delta_pivotal = utils.pivotal_ci_t(
            t_score_glass, df2, sample_size_control, confidence_level
        )
        (
            ci_lower_glass_delta_central,
            ci_upper_glass_delta_central,
            standard_error_glass_delta,
        ) = utils.ci_from_cohens_paired(
            glass_delta, sample_size_control, confidence_level
        )
        (
            ci_lower_glass_delta_unbiased_central,
            ci_upper_glass_delta_unbiased_central,
            standard_error_glass_delta_unbiased,
        ) = utils.ci_from_cohens_paired(
            unbiased_glass_delta, sample_size_control, confidence_level
        )

        inferential: res.InferentialStatistics = res.InferentialStatistics(
            score=round(t_score, 4),
            p_value=round(p_value, 4),
        )
        inferential.degrees_of_freedom = round(df, 4)
        inferential.standard_error = round(standard_error, 4)
        inferential.means_difference = round(sample_mean_difference, 4)

        glass_delta_es: es.GlassDelta = es.GlassDelta(
            value=round(glass_delta, 4),
            ci_lower=round(ci_lower_glass_delta_central, 4),
            ci_upper=round(ci_upper_glass_delta_central, 4),
            standard_error=round(standard_error_glass_delta, 4),
        )
        glass_delta_es.standardizer = round(standardizer_glass_delta, 4)
        glass_delta_es.update_pivotal_ci(
            round(ci_lower_glass_delta_pivotal, 4),
            round(ci_upper_glass_delta_pivotal, 4),
        )

        glass_delta_unbiased: es.GlassDeltaUnbiased = es.GlassDeltaUnbiased(
            value=round(unbiased_glass_delta, 4),
            ci_lower=round(ci_lower_glass_delta_unbiased_central, 4),
            ci_upper=round(ci_upper_glass_delta_unbiased_central, 4),
            standard_error=round(standard_error_glass_delta_unbiased, 4),
        )
        glass_delta_unbiased.standardizer = round(standardizer_unbiased_glass_delta, 4)
        glass_delta_unbiased.update_pivotal_ci(
            round(ci_lower_glass_delta_pivotal * correction, 4),
            round(ci_upper_glass_delta_pivotal * correction, 4),
        )

        results = TwoIndependentControlGroupResults()
        results.inferential = inferential
        results.glass_delta = glass_delta_es
        results.glass_delta_unbiased = glass_delta_unbiased

        return results

    @staticmethod
    def from_data(
        columns: list[list], population_mean_diff: float, confidence_level: float = 0.95
    ) -> TwoIndependentControlGroupResults:
        """
        Creates test results from the given data.

        Args:
            columns (list[list]): The data columns for the control and experimental groups.
            population_mean_diff (float): The population mean difference.
            confidence_level (float): The confidence level for the tests. Defaults to 0.95.

        Returns:
            TwoIndependentControlGroupResults: The results of the tests.

        Raises:
            ValueError: If columns length is not 2.
        """
        if len(columns) != 2:
            raise ValueError(texts.Errors.columns_must_be_two)

        column_1 = columns[0]
        column_2 = columns[1]

        sample_mean_control = np.mean(column_1)
        sample_mean_experimental = np.mean(column_2)
        sample_sd_control = np.std(column_1, ddof=1)
        sample_sd_experimental = np.std(column_2, ddof=1)
        sample_size_control = len(column_1)
        sample_size_experimental = len(column_2)
        sample_size = sample_size_control + sample_size_experimental
        df = sample_size - 2
        sample_mean_difference = sample_mean_experimental - sample_mean_control
        standardizer_glass_delta = np.sqrt(
            (
                (((sample_size_experimental - 1) * sample_sd_experimental**2))
                + ((sample_size_control - 1) * sample_sd_control**2)
            )
            / (sample_size - 2)
        )
        standard_error = standardizer_glass_delta * np.sqrt(
            (sample_size_experimental + sample_size_control)
            / (sample_size_experimental * sample_size_control)
        )
        t_score = (sample_mean_difference - population_mean_diff) / standard_error
        t_score_glass = (sample_mean_difference - population_mean_diff) / (
            sample_sd_control / np.sqrt(sample_size_control)
        )
        p_value = min(float(stats.t.sf((abs(t_score)), df) * 2), 0.99999)
        glass_delta = (
            sample_mean_difference - population_mean_diff
        ) / sample_sd_control
        df2 = sample_size_control - 1
        correction = math.exp(
            math.lgamma(df2 / 2)
            - math.log(math.sqrt(df2 / 2))
            - math.lgamma((df2 - 1) / 2)
        )
        unbiased_glass_delta = glass_delta * correction
        standardizer_unbiased_glass_delta = standardizer_glass_delta / correction
        ci_lower_glass_delta_pivotal, ci_upper_glass_delta_pivotal = utils.pivotal_ci_t(
            t_score_glass, df2, sample_size_control, confidence_level
        )
        (
            ci_lower_glass_delta_central,
            ci_upper_glass_delta_central,
            standard_error_glass_delta,
        ) = utils.ci_from_cohens_paired(
            float(glass_delta), sample_size_control, confidence_level
        )
        (
            ci_lower_glass_delta_unbiased_central,
            ci_upper_glass_delta_unbiased_central,
            standard_error_glass_delta_unbiased,
        ) = utils.ci_from_cohens_paired(
            float(unbiased_glass_delta), sample_size_control, confidence_level
        )

        experimental_group: res.Group = res.Group(
            mean=float(sample_mean_experimental),
            standard_deviation=float(sample_sd_experimental),
        )
        experimental_group.sample_size = sample_size_experimental
        experimental_group.mean_diff = float(round(sample_mean_difference, 4))

        control_group: res.Group = res.Group(
            mean=float(sample_mean_control),
            standard_deviation=float(sample_sd_control),
        )
        control_group.sample_size = sample_size_control
        control_group.mean_diff = float(round(sample_mean_difference, 4))

        inferential: res.InferentialStatistics = res.InferentialStatistics(
            score=round(t_score, 4), p_value=round(p_value, 4)
        )
        inferential.degrees_of_freedom = round(df, 4)
        inferential.standard_error = round(standard_error, 4)

        glass_delta_es: es.GlassDelta = es.GlassDelta(
            value=round(float(glass_delta), 4),
            standard_error=round(standard_error_glass_delta, 4),
            ci_lower=round(ci_lower_glass_delta_central, 4),
            ci_upper=round(ci_upper_glass_delta_central, 4),
        )
        glass_delta_es.standardizer = round(standardizer_glass_delta, 4)
        glass_delta_es.update_pivotal_ci(
            round(ci_lower_glass_delta_pivotal, 4),
            round(ci_upper_glass_delta_pivotal, 4),
        )

        glass_delta_unbiased_es: es.GlassDeltaUnbiased = es.GlassDeltaUnbiased(
            value=round(float(unbiased_glass_delta), 4),
            standard_error=round(standard_error_glass_delta_unbiased, 4),
            ci_lower=round(ci_lower_glass_delta_unbiased_central, 4),
            ci_upper=round(ci_upper_glass_delta_unbiased_central, 4),
        )
        glass_delta_unbiased_es.standardizer = round(
            standardizer_unbiased_glass_delta, 4
        )
        glass_delta_unbiased_es.update_pivotal_ci(
            round(ci_lower_glass_delta_pivotal * correction, 4),
            round(ci_upper_glass_delta_pivotal * correction, 4),
        )

        results = TwoIndependentControlGroupResults()
        results.experimental_group = experimental_group
        results.control_group = control_group
        results.inferential = inferential
        results.glass_delta = glass_delta_es
        results.glass_delta_unbiased = glass_delta_unbiased_es

        return results



from src.esek.Calculator.TwoIndependentMean.two_independent_control_group import (
    TwoIndependentControlGroupResults,
    TwoIndependentControlGroupTests,
)


def test_two_independent_control_group_from_score():
    assert True, "Method not implemented yet"


def test_two_independent_control_group_from_parameters():
    data = {
        "sample_mean_experimental": 5.0,
        "sample_mean_control": 4.0,
        "sample_sd_experimental": 1.0,
        "sample_sd_control": 1.0,
        "sample_size_experimental": 30,
        "sample_size_control": 30,
        "population_mean_diff": 0,
        "confidence_level": 0.95,
    }
    try:
        results = TwoIndependentControlGroupTests.from_parameters(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, TwoIndependentControlGroupResults)


def test_two_independent_control_group_from_data():
    data = {
        "columns": [[1, 2, 3, 4], [5, 6, 7, 8]],
        "population_mean_diff": 0,
        "confidence_level": 0.95,
    }
    try:
        results = TwoIndependentControlGroupTests.from_data(**data)
    except Exception as e:
        assert False, f"Unexpected error: {e}"

    assert isinstance(results, TwoIndependentControlGroupResults)



from setuptools import setup, find_packages

setup(
    name="esek",
    use_scm_version=True,
    setup_requires=["setuptools-scm"],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
)



"""
TwoIndependentAparametric module provides functionality for performing two independent aparametric tests.
It includes methods for calculating statistics such as U-statistic, mean ranks, and rank-biserial correlation.
The results are encapsulated in a dataclass for easy access and manipulation.
"""

from dataclasses import dataclass
from typing import Optional
from collections import Counter
import numpy as np
from scipy import stats
from ...utils import interfaces, res, utils, es, texts


@dataclass
class TwoIndependentAparametricResults:
    """
    A class to store results from two independent aparametric tests.
    """

    group1: Optional[res.Group] = None
    group2: Optional[res.Group] = None
    exact_inferential: Optional[res.InferentialStatistics] = None
    inferential: Optional[res.InferentialStatistics] = None
    corrected_inferential: Optional[res.InferentialStatistics] = None
    rank_biserial_correlation: Optional[es.Biserial] = None
    z_based_rank_biserial_correlation: Optional[es.Biserial] = None
    z_based_rank_biserial_correlation_corrected: Optional[es.Biserial] = None


class TwoIndependentAparametricTests(interfaces.AbstractTest):
    """
    A class to perform two independent aparametric tests.
    This class implements methods to calculate various statistics and effect sizes.

    It is designed to be used with two independent samples, calculating U-statistic, mean ranks,
    and rank-biserial correlation, among other statistics.

    Attributes:
        None

    Methods:
        from_score: Not implemented.
        from_parameters: Not implemented.
        from_data: Creates results from data columns.
        compute_confidence_interval: Computes confidence intervals for a given correlation.
    """

    @staticmethod
    def from_score() -> TwoIndependentAparametricResults:
        """
        A static method to create results from a score.
        This method is not implemented and will raise a NotImplementedError.
        Raises:
            NotImplementedError: This method is not implemented for TwoIndependentAparametric.
        """
        raise NotImplementedError(
            "from_score method is not implemented for TwoIndependentAparametric."
        )

    @staticmethod
    def from_parameters() -> TwoIndependentAparametricResults:
        """
        A static method to create results from parameters.
        This method is not implemented and will raise a NotImplementedError.
        Raises:
            NotImplementedError: This method is not implemented for TwoIndependentAparametric.
        """
        raise NotImplementedError(
            "from_parameters method is not implemented for TwoIndependentAparametric."
        )

    @staticmethod
    def from_data(
        columns: list, confidence_level: float = 0.95
    ) -> TwoIndependentAparametricResults:
        """
        Create results from data columns.

        Args:
            columns (list): A list containing two numpy arrays or lists representing the two independent samples.
            confidence_level (float): The confidence level for the confidence intervals, default is 0.95.
        Returns:
            TwoIndependentAparametricResults: An instance containing the results of the two independent aparametric tests.
        Raises:
            ValueError: If the length of columns is not 2.
        """
        if len(columns) != 2:
            raise ValueError(texts.Errors.columns_must_be_two)

        column_1 = columns[0]
        column_2 = columns[1]
        sample_median_1 = np.median(column_1)
        sample_median_2 = np.median(column_2)
        sample_mean_1 = np.mean(column_1)
        sample_mean_2 = np.mean(column_2)
        sample_size_1 = len(column_1)
        sample_size_2 = len(column_2)
        sample_size = sample_size_1 + sample_size_2

        merged_samples = np.append(column_1, column_2)
        ranks = stats.rankdata(merged_samples, method="average")
        sum_ranks_1 = np.sum(ranks[:sample_size_1])
        sum_ranks_2 = np.sum(ranks[sample_size_1:])
        mean_ranks_1 = np.mean(ranks[:sample_size_1])
        mean_ranks_2 = np.mean(ranks[sample_size_1:])

        freq = Counter(merged_samples)
        frequencies = list(freq.values())
        ties_correction = [(f**3 - f) for f in frequencies]
        multiplicity_factor = sum(ties_correction)

        mean_w_1 = (sample_size_1 * (sample_size + 1)) / 2
        mean_w_2 = (sample_size_2 * (sample_size + 1)) / 2
        u_statistic_1 = sum_ranks_1 - (sample_size_1 * (sample_size_1 + 1)) / 2
        u_statistic_2 = sum_ranks_2 - (sample_size_2 * (sample_size_2 + 1)) / 2
        variance = (
            (sample_size_1 * sample_size_2 * (sample_size_1 + sample_size_2 + 1)) / 12
        ) - ((sample_size_1 * sample_size_2 * (multiplicity_factor))) / (
            12 * sample_size * (sample_size - 1)
        )

        z_score = abs(sum_ranks_1 - mean_w_1) / np.sqrt(variance)
        z_score_corrected = (abs(sum_ranks_1 - mean_w_1) - 0.5) / np.sqrt(variance)

        p_value = min(float(stats.norm.sf((abs(z_score))) * 2), 0.99999)
        p_value_corrected = min(
            float(stats.norm.sf((abs(z_score_corrected))) * 2), 0.99999
        )

        rank_biserial_correlation = (u_statistic_1 - u_statistic_2) / (
            u_statistic_1 + u_statistic_2
        )
        rosenthal_rank_biserial_correlation_z = z_score / np.sqrt(sample_size)
        rosenthal_rank_biserial_correlation_z_corrected = z_score_corrected / np.sqrt(
            sample_size
        )

        exact_p_value = stats.mannwhitneyu(column_1, column_2, method="exact")[1]

        standard_error_rbc = np.sqrt(
            (sample_size_1 + sample_size_2 + 1) / (3 * sample_size_1 + sample_size_2)
        )
        z_critical_value = float(
            stats.norm.ppf((1 - confidence_level) + ((confidence_level) / 2))
        )

        lower_ci_rank_biserial_correlation, upper_ci_rank_biserial_correlation = (
            utils.compute_fisher_confidence_interval(
                rank_biserial_correlation, standard_error_rbc, (z_critical_value)
            )
        )

        lower_ci_z_based, upper_ci_z_based = utils.compute_fisher_confidence_interval(
            rosenthal_rank_biserial_correlation_z, standard_error_rbc, z_critical_value
        )

        lower_ci_z_based_corrected, upper_ci_z_based_corrected = (
            utils.compute_fisher_confidence_interval(
                rosenthal_rank_biserial_correlation_z_corrected,
                standard_error_rbc,
                z_critical_value,
            )
        )

        group1 = res.Group(
            mean=round(mean_w_1, 4),
            median=round(sample_median_1, 4),
            standard_deviation=round(np.sqrt(variance), 4),
        )
        group2 = res.Group(
            mean=round(mean_w_2, 4),
            median=round(sample_median_2, 4),
            standard_deviation=round(np.sqrt(variance), 4),
        )
        group1.sample_size = sample_size_1
        group1.u_statistic = round(u_statistic_1, 4)
        group1.w_statistic = round(sum_ranks_1, 4)
        group1.mean_rank = float(round(mean_ranks_1, 4))

        group2.sample_size = sample_size_2
        group2.u_statistic = round(u_statistic_2, 4)
        group2.w_statistic = round(sum_ranks_2, 4)
        group2.mean_rank = float(round(mean_ranks_2, 4))

        inferential = res.InferentialStatistics(
            score=round(z_score, 4),
            p_value=round(p_value, 6),
        )
        corrected_inferential = res.InferentialStatistics(
            score=round(z_score_corrected, 4),
            p_value=round(p_value_corrected, 6),
        )
        exact_inferential = res.InferentialStatistics(
            score=round(z_score, 4),
            p_value=round(exact_p_value, 6),
        )

        rank_biserial_correlation_result = es.Biserial(
            name="Rank Biserial Correlation",
            value=round(rank_biserial_correlation, 4),
            ci_lower=lower_ci_rank_biserial_correlation,
            ci_upper=upper_ci_rank_biserial_correlation,
            standard_error=standard_error_rbc,
        )
        z_based_rank_biserial_correlation_result = es.Biserial(
            name="Z Based Rank Biserial Correlation",
            value=round(rosenthal_rank_biserial_correlation_z, 4),
            ci_lower=lower_ci_z_based,
            ci_upper=upper_ci_z_based,
            standard_error=standard_error_rbc,
        )
        z_based_rank_biserial_correlation_corrected_result = es.Biserial(
            name="Z Based Rank Biserial Correlation (Corrected)",
            value=round(rosenthal_rank_biserial_correlation_z_corrected, 6),
            ci_lower=lower_ci_z_based_corrected,
            ci_upper=upper_ci_z_based_corrected,
            standard_error=standard_error_rbc,
        )

        results = TwoIndependentAparametricResults()
        results.group1 = group1
        results.group2 = group2
        results.inferential = inferential
        results.corrected_inferential = corrected_inferential
        results.exact_inferential = exact_inferential
        results.rank_biserial_correlation = rank_biserial_correlation_result
        results.z_based_rank_biserial_correlation = (
            z_based_rank_biserial_correlation_result
        )
        results.z_based_rank_biserial_correlation_corrected = (
            z_based_rank_biserial_correlation_corrected_result
        )

        return results
